#
# Copyright (c) 2020, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
import json
import re
import requests
from gettext import gettext as _
from os import system as os_system, remove as os_remove
from os.path import join as path_join, exists, basename
from wave import open as wave_open

from astersay.utils import make_checksum
from astersay.backends.yandex.base import YandexBaseBackend


class YandexSynthesizer(YandexBaseBackend):
    URL = 'https://tts.api.cloud.yandex.net/speech/v1/tts:synthesize'
    LANGS = ('ru-RU', 'en-US', 'tr-TR')
    VOICES = {
        'ru-RU': ('oksana', 'jane', 'omazh', 'zahar', 'ermil',
                  'alena', 'filipp'),
        'tr-TR': ('silaerkan', 'erkanyavas'),
        'en-US': ('alyss', 'nick'),
    }
    EMOTIONS = ('good', 'evil', 'neutral')
    AUDIO_FORMATS = ('lpcm', 'oggopus')
    HERTZ_LIST = (48000, 16000, 8000)

    def __init__(self, settings):
        YandexBaseBackend.__init__(self)
        self.set_token_manager(settings)

        conf = settings.yandex
        self.folder_id = conf['auth']['folder_id']

        synthesizer = conf['synthesizer']
        self.storage_path = synthesizer['storage']
        self.convertor = synthesizer['convertor']

        lang = synthesizer['lang']
        assert lang in self.LANGS
        self.lang = lang

        voice = synthesizer['voice']
        assert voice in self.VOICES[lang]
        self.voice = voice

        emotion = synthesizer['emotion']
        assert emotion in self.EMOTIONS
        self.emotion = emotion

        speed = synthesizer['speed']
        assert 0.1 <= speed <= 3.0
        self.speed = speed

        audio_format = synthesizer['format']
        assert audio_format in self.AUDIO_FORMATS
        self.audio_format = audio_format

        sample_rate_hertz = synthesizer['sample_rate_hertz']
        assert sample_rate_hertz in self.HERTZ_LIST
        self.sample_rate_hertz = sample_rate_hertz

    @property
    def message_params(self):
        return {
            'lang': self.lang,
            'voice': self.voice,
            'emotion': self.emotion,
            'speed': self.speed,
            'folderId': self.folder_id,
            'format': self.audio_format,
            'sampleRateHertz': self.sample_rate_hertz,
        }

    def make_message(self, text):
        assert len(text) <= 5000
        message = self.message_params
        message['text'] = text
        return message

    def _synthesize(self, message: dict):
        self.token_manager.update_token()
        logger_debug = self.logger.debug

        headers = {
            'Authorization': 'Bearer %s' % self.token_manager.token,
        }
        logger_debug(_('Заголовки синтеза речи: %s'), headers)
        logger_debug(_('Параметры синтеза речи: %s'), message)
        response = requests.post(self.URL, data=message, headers=headers)
        if response.ok:
            logger_debug(_('Синтез речи выполнен.'))
            return response.content
        self.logger.warning(
            _('Ошибка синтеза речи. Яндекс вернул код, отличный от 200. %s'),
            response.text,
        )

    def make_voice(self, text, early=False):
        logger_debug = self.logger.debug
        logger_info = self.logger.info
        logger_info(_('Начинается синтез речи.'))

        # Убираем лишние пробелы.
        text = re.sub(r'\s+', ' ', text)

        convertor = self.convertor

        def find_voice(checksum):
            voice = path_join(self.storage_path, checksum)
            filename = '%s.wav' % voice
            # Именно такой файл уже есть, не нужно синтезировать.
            if exists(filename):
                logger_info(
                    _('Речь %r существует и будет взята из хранилища.'), voice)
                found = True
            else:
                found = False
            return voice, filename, found

        def save_info(voice, spec):
            spec_filename = '%s.json' % voice
            json.dump(spec, open(spec_filename, 'w'), indent=4, ensure_ascii=False)
            logger_debug(_('Сохранён файл спецификации: %s'), spec_filename)

        def synthesize(voice, filename, message, spec):
            save_info(voice, spec)

            audio_data = self._synthesize(message)

            src_filename = '%s.%s' % (voice, convertor['src_format'])
            with open(src_filename, 'wb') as f:
                f.write(audio_data)
            logger_info(_('Записан исходный файл речи: %s'), src_filename)
            command = convertor['command'] % (src_filename, filename)
            os_system(command)
            logger_info(_('Записан конечный файл речи: %s'), filename)

            if convertor.get('src_delete'):
                os_remove(src_filename)
            return True

        # Символ | является разделителем на несколько составных файлов, но
        # он может быть и цельным.
        message = self.make_message(text.replace('|', ''))
        spec = message.copy()
        spec['convertor'] = convertor
        checksum = make_checksum(spec)
        voice, filename, found = find_voice(checksum)

        texts = [x for x in text.split('|') if x.strip()]
        if len(texts) == 1:
            if not found:
                logger_info(_('Синтезирую в речь текст: %r'), text)
                synthesize(voice, filename, message, spec)
        elif not found:
            chunks = []
            for chunk_text in texts:
                chunk_message = self.make_message(chunk_text)
                chunk_spec = chunk_message.copy()
                chunk_spec['convertor'] = convertor
                chunk_checksum = make_checksum(chunk_spec)
                chunk_voice, chunk_filename, found = find_voice(chunk_checksum)
                if not found:
                    logger_info(_('Синтезирую в речь текст: %r'), chunk_text)
                    synthesize(chunk_voice, chunk_filename, chunk_message,
                               chunk_spec)
                chunks.append((chunk_voice, chunk_filename))

            # Если это заблаговременный синтез частей текста, то не собираем
            # конечный файл, а только все части.
            if early:
                logger_debug(_('Для предварительного синтеза речи конечный '
                               'файл не собирается.'))
                return voice

            frames = []
            params = None
            last_voice = chunks[-1][0]
            for chunk_voice, chunk_filename in chunks:
                with wave_open(chunk_filename) as f:
                    if chunk_voice == last_voice:
                        params = f.getparams()
                    frames.append(f.readframes(f.getnframes()))
            # Записываем информацию по слитому файлу.
            spec['chunks'] = [basename(v) for v, f in chunks]
            save_info(voice, spec)
            # Записываем конечный, слитый файл.
            with wave_open(filename, 'wb') as f:
                f.setparams(params)
                for frame in frames:
                    f.writeframes(frame)

        return voice



#
# Copyright (c) 2020, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
import pyaudio
import sys
import wave
from gettext import gettext as _
from os.path import exists
from threading import Thread
from time import time, sleep


class StreamIO:
    mode = '+'

    def __init__(self):
        self.data = []

    def write(self, content):
        self.data.append(content)

    def read(self, size=-1):
        raise NotImplementedError

    def readline(self, size=-1):
        pause = 0.2
        total = 30 / pause
        n = 0
        data = self.data
        while not data and n < total:
            sleep(pause)
            n += 1
        if not data:
            return ''
        return data.pop(0)

    def flush(self):
        pass

    def close(self):
        pass


class DummyAsterisk:

    AUDIO_CHUNK_SIZE = 1024
    READ_COMMAND_TIMEOUT = 0.2
    commands = (
        'GET VARIABLE', 'SET VARIABLE', 'VERBOSE', 'STREAM FILE', 'EXEC')
    stream = None
    is_stopped = False

    def __init__(self, agi_request=None, model='', workdir='',
                 stdin=None, stdout=None):
        if stdin is None:
            stdin = StreamIO()
        if stdout is None:
            stdout = StreamIO()
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = sys.stderr

        agi_args = [
            ('agi_request', agi_request or __file__),
            ('agi_channel', 'PJSIP/001-00000001'),
            ('agi_language', 'ru'),
            ('agi_type', 'PJSIP'),
            ('agi_uniqueid', str(round(time(), 2))),
            ('agi_version', '0.0.1'),
            ('agi_callerid', '001'),
            ('agi_calleridname', 'test'),
            ('agi_callingpres', '0'),
            ('agi_callingani2', '0'),
            ('agi_callington', '0'),
            ('agi_callingtns', '0'),
            ('agi_dnid', '1000'),
            ('agi_rdnis', 'unknown'),
            ('agi_context', 'from-internal'),
            ('agi_extension', '1000'),
            ('agi_priority', '4'),
            ('agi_enhanced', '1.0'),
            ('agi_accountcode', ''),
            ('agi_threadid', id(self)),
        ]
        if workdir:
            agi_args += [
                ('agi_arg_1', model),
                ('agi_arg_2', workdir),
            ]
        elif model:
            agi_args += [
                ('agi_arg_1', model),
            ]
        write = self.stdout.write
        for name, value in agi_args:
            # Передаём приложению.
            write('%s:%s\n' % (name, value))
        write('\n')
        self.stdout.flush()

        self.agi_args = agi_args
        self.variables = {}
        self.playlist = []

        self.audio = pyaudio.PyAudio()
        self._playback_process = None
        self._playback_args = None
        self._playback_cancel = False

    def parse_command(self, line):
        for command in self.commands:
            cmd = command + ' '
            if line.startswith(cmd) or line.startswith(cmd.lower()):
                method_name = 'command_' + command.replace(' ', '_').lower()
                method = getattr(self, method_name, None)
                if method:
                    start_string = False
                    args = []
                    for s in line[len(cmd):].split(' '):
                        start_quote = s.startswith('"')
                        end_quote = s.endswith('"')
                        if start_quote and end_quote and s != '"':
                            args.append(s.strip('"'))
                            start_string = False
                        elif start_quote and not start_string:
                            args.append(s[1:])
                            start_string = True
                        elif end_quote and start_string:
                            args[-1] += ' ' + s[:-1]
                            start_string = False
                        elif start_string:
                            args[-1] += ' ' + s
                        else:
                            args.append(s)
                    return method, args
        return None, ()

    def send_result(self, result, code=200):
        msg = '%d result=%s' % (code, result)
        self.stdout.write(msg + '\n')
        self.stdout.flush()
        # self.display('self.stdout.write(%r)' % msg)

    def display(self, *messages):
        write = self.stderr.write
        for m in messages:
            write(m + '\n')
        self.stderr.flush()

    def stop_record_audio(self):
        if self.stream:
            self.stream.stop_stream()
            self.stream.close()
            msg = _('Остановлен поток записи: %s') % self.stream
            self.display(msg)

    def start_record_audio(self):
        self.stop_record_audio()
        self.stream = self.audio.open(
            format=pyaudio.paInt16,
            rate=8000,
            channels=1,
            input=True,
            frames_per_buffer=self.AUDIO_CHUNK_SIZE
        )
        msg = _('Создан поток записи: %s') % self.stream
        self.display(msg)

    def stop_playback(self, clear=False, digit='0'):
        if clear:
            self.playlist.clear()
        self._playback_cancel = digit
        # Дожидаемся отключения процесса.
        while self._playback_process:
            self.display(_('Дожидаемся отключения старого процесса '
                           'проигрывания аудио.'))
            sleep(0.5)
            if not self._playback_process:
                self.display(_('Cтарый процесс проигрывания аудио завершён.'))
        # Оставшийся плейлист продолжается.
        if self.playlist:
            self.start_playback()
        return True

    def start_playback(self):

        def playback():
            self.display(_('Начинается плейлист.'))
            while self.playlist:
                filename, args = self.playlist.pop(0)
                self.display(
                    _('Проигрываем %s с агрументами %s') % (filename, args))
                self._playback_args = args
                try:
                    wfile = wave.open(filename, 'rb')
                except IOError:
                    self.display(_('Файл %s не найден.') % filename)
                    continue
                p = self.audio
                try:
                    stream = p.open(
                        format=p.get_format_from_width(wfile.getsampwidth()),
                        channels=wfile.getnchannels(),
                        rate=wfile.getframerate(),
                        output=True
                    )
                    chunk_size = self.AUDIO_CHUNK_SIZE
                    data = wfile.readframes(chunk_size)
                    while data and not self._playback_cancel:
                        stream.write(data)
                        data = wfile.readframes(chunk_size)
                    if not self._playback_cancel:
                        # Когда воспроизведение не было прервано.
                        self.variables['PLAYBACKSTATUS'] = 'SUCCESS'
                        self.send_result('0')
                    else:
                        self.send_result(self._playback_cancel)
                        self.display(_('Воспроизведение прервано.'))
                    sleep(0.5)
                    stream.stop_stream()
                    stream.close()
                except Exception as e:
                    self.display(str(e))
            self._playback_process = None
            self._playback_args = None
            self.display(_('Плейлист окончен.'))

        if self._playback_process is None:
            self._playback_args = None
            self._playback_cancel = False
            self._playback_process = Thread(target=playback, daemon=True)
            self._playback_process.start()

    def play_audio(self, filename, *args, clear=False):
        if clear:
            self.stop_playback(clear)
        if exists(filename):
            pass
        elif exists(filename + '.wav'):
            filename += '.wav'
        # elif exists(filename + '.ogg'):
        #     filename += '.ogg'
        self.playlist.append([filename, args])
        self.start_playback()

    def start(self):
        readline = self.stdin.readline
        parse_command = self.parse_command
        while True:
            if self.is_stopped:
                break
            try:
                line = readline().strip()
            except KeyboardInterrupt:
                self.stop()
                break
            # self.display('line=%r' % line)
            if line and line in ('exit', 'quit'):
                self.stop()
                break
            elif line:
                method, args = parse_command(line)
                if method:
                    try:
                        method(*args)
                    except Exception as e:
                        self.display(str(e))
                else:
                    self.display(_('Метод не найден.'))
            else:
                sleep(self.READ_COMMAND_TIMEOUT)
        self.display(_('Сервер остановлен.'))

    def stop(self):
        self.display(_('Сервер останавливается.'))
        self.is_stopped = True
        # self.stdin.write('\n')
        self.stop_record_audio()
        self.stop_playback(clear=True)
        self.stdin.close()
        self.stdout.close()
        self.stderr.close()
        self.audio.terminate()
        self.display(_('Сервер остановлен.'))

    def command_get_variable(self, name: str):
        """
        GET VARIABLE VARIABLENAME
        Returns 0 if variablename is not set. Returns 1 if variablename is set
        and returns the variable in parentheses. Example return code:
        200 result=1 (testvariable)
        """
        if name not in self.variables:
            value = '0'
        else:
            value = '1 (%s)' % self.variables[name]
        self.send_result(value)

    def command_set_variable(self, name: str, value: str):
        """
        SET VARIABLE VARIABLENAME VALUE
        Sets a variable to the current channel.
        """
        self.variables[name] = value
        self.display('command_set_variable(%r, %r)' % (name, value))

    def command_verbose(self, message: str, level: str):
        """
        VERBOSE MESSAGE LEVEL
        Sends message to the console via verbose message system. Level is the
        verbose level (1-4). Always returns 1.
        """
        if level not in ('1', '2', '3', '4'):
            raise ValueError(
                _('Уровень должен быть от 1 до 4, level == %r') % level)
        self.display('command_verbose(%r, %r)' % (message, level))

    def command_stream_file(self, filename: str, escape_digits: str, sample_offset: str):
        """
        STREAM FILE FILENAME ESCAPE_DIGITS SAMPLE OFFSET
        Send the given file, allowing playback to be interrupted by the given
        digits, if any. Returns 0 if playback completes without a digit being
        pressed, or the ASCII numerical value of the digit if one was pressed,
        or -1 on error or if the channel was disconnected. If musiconhold is
        playing before calling stream file it will be automatically stopped
        and will not be restarted after completion.
        It sets the following channel variables upon completion:
        PLAYBACKSTATUS - The status of the playback attempt as a text string.
        SUCCESS
        FAILED
        """
        if escape_digits and not escape_digits.isdigit():
            raise ValueError(
                _('escape_digits должны быть цифрами, а не %r') % escape_digits)
        self.play_audio(filename, escape_digits, sample_offset, clear=True)
        self.display('command_stream_file(%r, %r, %r)' % (
            filename, escape_digits, sample_offset))

    def command_exec(self, application: str, options: str):
        """
        EXEC APPLICATION OPTIONS
        Executes application with given options.
        Returns whatever the application returns, or -2 on failure to
        find application.
        """
        self.display('command_exec(%r, %r)' % (application, options))
        if application.lower() == 'senddtmf':
            opts = [s.strip() for s in options.split(',')]
            if opts and opts[0].isdigit():
                digit = opts[0][0]
                args = self._playback_args
                if args and digit in args[0]:
                    self.display('_playback_args: %s' % str(args))
                    # Останавливаем текущее воспроизведение.
                    self.stop_playback(digit)
                    self.display(_('Отправлено нажатие %s.') % digit)
                return
        self.send_result('-2')


if __name__ == "__main__":
    server = DummyAsterisk(stdin=sys.stdin, stdout=sys.stdout)
    server.start()



#
# Copyright (c) 2020, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
import json
import jwt
import logging
import os
import requests
from dateutil.parser import parse as parse_datetime
from gettext import gettext as _
from time import time

from astersay.exceptions import ImproperlyConfigured, TokenError

logger = logging.getLogger(__name__)


class YandexIamToken:

    @classmethod
    def from_response(cls, jsondata):
        key = jsondata['iamToken']
        expiration = parse_datetime(jsondata['expiresAt']).timestamp()
        return cls(key, expiration)

    def __init__(self, key, expiration):
        self.key = str(key)
        self.expiration = float(expiration)

    def __str__(self):
        return self.key or 'invalid'

    def expires_after(self, seconds=0):
        """Проверяет истечение токена через переданное кол-во секунд."""
        return time() + seconds >= self.expiration

    @property
    def is_expired(self):
        return self.expires_after(0)

    @property
    def is_valid(self):
        return len(self.key) >= 512


class YandexTokenManager:

    def __init__(self, private_filename, token_filename, key_id,
                 service_account_id, **kwargs):
        self.private_filename = private_filename
        self.token_filename = token_filename
        self.key_id = key_id
        self.service_account_id = service_account_id

        save = False
        if os.path.exists(token_filename):
            try:
                data = json.load(open(token_filename))
                self.token = YandexIamToken(data['key'], data['expiration'])
            except (ValueError, KeyError):
                self.token = YandexIamToken('', 0)
                save = True
        else:
            self.token = YandexIamToken('', 0)
            save = True
        if save:
            data = self.save_token()
            json.dump(data, open(token_filename, 'w'), indent=4)

    def save_token(self):
        t = self.token
        data = {'key': t.key, 'expiration': t.expiration}
        json.dump(data, open(self.token_filename, 'w'), indent=4)
        logger.info(_('IAM-токен сохранён в файл: %s'), self.token_filename)
        return data

    def make_json_web_token(self, lifetime=360):
        """
        Создаёт новый JWT с помощью которого будет периодически запрашиваться
        новый IAM-токен. Время жизни этого токена по умолчанию - 10 минут.
        """
        if not self.service_account_id:
            raise ImproperlyConfigured(
                _('Для Яндекс не установлен service_account_id.'))
        if not self.key_id:
            raise ImproperlyConfigured(
                _('Для Яндекс не установлен key_id.'))

        with open(self.private_filename, 'r') as private:
            private_key = private.read()

        now = int(time())
        payload = {
            'aud': 'https://iam.api.cloud.yandex.net/iam/v1/tokens',
            'iss': self.service_account_id,
            'iat': now,
            'exp': now + lifetime
        }
        token = jwt.encode(
            payload,
            private_key,
            algorithm='PS256',
            headers={'kid': self.key_id},
        )
        if isinstance(token, bytes):
            token = token.decode('utf-8')
        logger.info(_('Создан новый JWT-токен.'))
        return token

    def get_new_iam_token(self, jwt_token=None):
        """
        Запрашивает новый IAM-токен у Яндекса.
        """
        if not jwt_token:
            jwt_token = self.make_json_web_token()
        logger.info(_('Запрашивается новый IAM-токен.'))
        response = requests.post(
            "https://iam.api.cloud.yandex.net/iam/v1/tokens",
            headers={"Content-Type": "application/json"},
            data=json.dumps({'jwt': jwt_token})
        )

        if response.status_code != 200:
            logger.error(_(
                'Ошибка запроса на обновление IAM-токена. '
                'Запрос вернул код, отличный от 200. %s'),
                response.content.decode('utf-8'))
            raise ConnectionError(
                _('Запрос вернул код %s.') % response.status_code)

        token = YandexIamToken.from_response(response.json())
        logger.debug(_('Новый IAM-токен успешно создан: %s'), token)
        return token

    def update_token(self):
        if not self.token.expires_after(360):
            logger.info(_('Время обновления IAM-токена ещё не подошло.'))
            return self.token
        try:
            self.token = self.get_new_iam_token()
            logger.info(_('IAM-токен обновлен.'))
            self.save_token()
        except ConnectionError as e:
            logger.critical(
                _('Ошибка во время обновления IAM-токена.'), exc_info=e)
        return self.token

    def validate(self):
        token = self.token
        if not token.is_expired:
            raise TokenError(_('IAM-токен устарел.'))
        if not token.is_valid:
            raise TokenError(_('IAM-токен недействителен.'))
        logger.debug(_('IAM-токен находится в актуальном состоянии.'))



#
# Copyright (c) 2020, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
import logging
import re
import signal
import sys
from gettext import gettext as _

from astersay.exceptions import (
    AgiUnknownError, AgiAppError, AgiHangupError, AgiSighupError,
    AgiSigpipeError, AgiUsageError, AgiInvalidCommand
)

logger = logging.getLogger(__name__)

re_code = re.compile(r'(^\d+)\s*(.*)')
re_kv = re.compile(r'(?P<key>\w+)=(?P<value>[^\s]+)\s*(?:\((?P<data>.*)\))*')


def quote_value(string):
    if not isinstance(string, str):
        string = str(string)
    if '"' in string:
        raise ValueError(
            _('Строка %r содержит двойную кавычку.') % string)
    return '"%s"' % string


def digits_value(digits):
    if not isinstance(digits, str):
        digits = ''.join(map(str, digits))
    elif digits and not digits.isdigit():
        digits = ''.join([x for x in digits if x in x.isdigit()])
    return quote_value(digits)


class Agi:
    is_sighup = False
    is_hungup = False

    def __init__(self, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr):
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        signal.signal(signal.SIGHUP, self._handle_sighup)
        # Входящие параметры.
        self.params = {}
        # Исходящие переменные.
        self.export = {}
        self._parse_params()

    def _handle_sighup(self, signum, frame):
        """Обработка сигнала SIGHUP."""
        self.is_sighup = True

    def _parse_params(self):
        """Обработка входных параметров."""
        readline = self.stdin.readline
        params = self.params
        # Заметьте, что здесь неуместно логирование, так как, его настройки
        # ещё не переданы и в этом месте оно происходит по дефолтным настройкам.
        # Это значит, что информация не попадёт в нужный лог.
        stderr_write = self.stderr.write
        while True:
            line = readline().strip()
            if not line:
                break
            if type(line) is bytes:
                line = line.decode('utf8')
            # stderr_write('line = %r\n' % line)
            if ':' not in line:
                error = _('В строке %r нарушен параметр.') % line
                stderr_write(error + '\n')
                continue
            param = line.split(':', 1)
            key = param[0].strip()
            data = param[1].strip()
            if key != '':
                params[key] = data
                # stderr_write('param %r = %r\n' % (key, data))
            else:
                error = _('Пропуск пустого параметра.')
                stderr_write(error + '\n')

    def get_result(self):
        """Считывает результат исполнения команды в Asterisk."""
        readline = self.stdin.readline
        line = readline().strip()
        if type(line) is bytes:
            line = line.decode('utf8')
        logger.debug('RESULT_LINE: %s', line)
        # Если ответа нет совсем, то отдаём пустой результат.
        if not line:
            return
        code = 0
        result = {'result': ('', '')}
        match = re_code.search(line)
        if match:
            code, response = match.groups()
            logger.debug('code=%s, response=%s', code, response)
            code = int(code)

        if code == 200:
            for key, value, data in re_kv.findall(response):
                result[key] = (value, data)
                # If user hangs up... we get 'hangup' in the data
                if data == 'hangup':
                    self.is_hungup = True
                    raise AgiHangupError(
                        _('Абонент повесил трубку во время исполнения.'))
                if key == 'result' and value == '-1':
                    raise AgiAppError(
                        _('Ошибка выполнения в Asterisk или сброс звонка.'))
            logger.debug('RESULT_DICT: %s', result)
            return result
        elif code == 510:
            raise AgiInvalidCommand(response)
        elif code == 520:
            usage = [line]
            line = readline().strip()
            if type(line) is bytes:
                line = line.decode('utf8')
            while line[:3] != '520':
                usage.append(line)
                line = readline().strip()
                if type(line) is bytes:
                    line = line.decode('utf8')
            usage.append(line)
            usage = '%s\n' % '\n'.join(usage)
            raise AgiUsageError(usage)
        raise AgiUnknownError(_('Неопределенный код ответа: %d.') % code)

    def send_command(self, command: str, *args):
        """Отправляет команду в Asterisk."""
        command = command.strip()
        command = '%s %s' % (command, ' '.join(map(str, args)))
        command = command.strip() + '\n'
        stdout = self.stdout
        if 'b' in stdout.mode:
            command = command.encode('utf-8')
        try:
            stdout.write(command)
            stdout.flush()
        except BrokenPipeError:
            raise AgiSigpipeError(_('Ошибка в конвейере (SIGPIPE).'))
        logger.debug(_('Отправлена команда %r'), command)
        return command

    def test_hangup(self):
        """Проверка завершения разговора со стороны пользователя."""
        if self.is_sighup:
            raise AgiSighupError(
                _('Из Asterisk пришёл сигнал SIGHUP.'))
        if self.is_hungup:
            raise AgiHangupError(
                _('Абонент повесил трубку.'))
        logger.debug(_('Проверка состояния разговора прошла успешно.'))

    def execute(self, command, *args):
        self.test_hangup()
        self.send_command(command, *args)

    def set(self, variable, value):
        """
        Отправляет в Asterisk переменную без проверки состояния конвейера.
        """
        self.export[variable] = value
        self.send_command('SET VARIABLE', variable, quote_value(value))

    def verbose(self, message, level=1):
        """
        Отправляет в Asterisk команду VERBOSE без проверки состояния конвейера.
        """
        self.send_command('VERBOSE', quote_value(message), level)

    def stream_file(self, filename, escape_digits='', sample_offset=0):
        """
        Отправляет файл, позволяя прервать его цифрами, если они переданы.
        Цифры- это строка или список цифр. Если предусмотрено смещение файла,
        то воспроизведение начнётся с этого места.
        Возвращает нажатую цифру или 0, если она не была нажата.
        """
        escape_digits = digits_value(escape_digits)
        self.execute(
            'STREAM FILE', filename, escape_digits, sample_offset)

    def exec_command(self, command, *args):
        """
        Отправляет произвольную команду Asterisk с аргументами (даже команды не
        для AGI).
        """
        args = ','.join(map(str, args))
        return self.execute('EXEC', command, args)



#!/usr/bin/env python3
"""
Пример плагина для Astresay. Первым аргументом для него должен выступать
путь к временному JSON-файлу с данными, полученными в процессе работы
диалога.
"""
import json
import sys

# Считываем файл с данными для плагина.
filename = sys.argv[1]
plugin_data = json.load(open(filename))

#######################
# Начало вашего кода. #
#######################


# Производим операции с ними.
print('Данные для плагина:', plugin_data)


######################
# Конец вашего кода. #
######################

# Обязательно указываем, что плагин успешно завершён, чтобы не прервать диалог.
sys.exit()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from .stt_service_pb2 import (
    StreamingRecognitionRequest, StreamingRecognitionResponse
)


class SttServiceStub(object):
    # missing associated documentation comment in .proto file
    pass

    def __init__(self, channel):
        """Constructor.

        Args:
          channel: A grpc.Channel.
        """
        self.StreamingRecognize = channel.stream_stream(
            '/yandex.cloud.ai.stt.v2.SttService/StreamingRecognize',
            request_serializer=StreamingRecognitionRequest.SerializeToString,
            response_deserializer=StreamingRecognitionResponse.FromString,
        )


class SttServiceServicer(object):
    # missing associated documentation comment in .proto file
    pass

    def StreamingRecognize(self, request_iterator, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_SttServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        'StreamingRecognize': grpc.stream_stream_rpc_method_handler(
            servicer.StreamingRecognize,
            request_deserializer=StreamingRecognitionRequest.FromString,
            response_serializer=StreamingRecognitionResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        'yandex.cloud.ai.stt.v2.SttService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))



#
# Copyright (c) 2020, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
"""
Функционал для обработки моделей диалогов.
"""
import json
import logging
import re
import subprocess
import wave
from collections import OrderedDict
from datetime import datetime
from gettext import gettext as _
from os.path import splitext, basename
from tempfile import NamedTemporaryFile
from time import time, sleep
from threading import Thread

from astersay.exceptions import AgiError, AgiHangupError, AgiSigpipeError

logger = logging.getLogger(__name__)
log_debug = logger.debug
log_info = logger.info
log_warning = logger.warning
log_error = logger.error


class TextBuffer(list):

    def __init__(self, key, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.key = key
        # Список индексов зафиксированных текстов.
        self.fixed = []
        # Время финализированного текста, должно сбрасываться предварительным
        # распознаванием.
        self.final = None

    def get_text(self, mode=1):
        """
        Возвращает текст буфера в режимах:
        1 - финальный, 2 - промежуточный, 3 - весь, 4 - весь с разметкой.
        """
        fixed = self.fixed
        if mode == 1:
            L = [x for i, x in enumerate(self) if x and i in fixed]
        elif mode == 2:
            L = [x for i, x in enumerate(self) if x and i not in fixed]
        elif mode == 3:
            L = [x for x in self if x]
        else:
            L = ['%s: %s' % (i + 1, x) for i, x in enumerate(self) if x]
        return '\n'.join(L)

    @property
    def seconds_after_final(self):
        if self.final:
            return time() - self.final
        return 0


def _make_say_params(saydata, variables):
    """Возвращает готовые аргументы для метода say()."""
    params = {'text': '', 'escape_digits': '', 'voice': ''}

    if 'voice' in saydata:
        params['voice'] = saydata['voice'] % variables
    elif 'file' in saydata:
        params['voice'] = splitext(saydata['file'])[0] % variables
    elif 'texts' in saydata:
        params['text'] = '|'.join(saydata['texts']) % variables
    else:
        params['text'] = saydata['text'] % variables

    if 'escape_digits' in saydata:
        escape_digits = saydata['escape_digits']
        if not isinstance(escape_digits, str):
            escape_digits = ''.join(map(str, escape_digits))
        params['escape_digits'] = escape_digits

    if 'nonblocking' in saydata:
        params['nonblocking'] = bool(saydata['nonblocking'])

    return params


def _make_pause(data, key):
    try:
        pause = float(data.get(key, 0))
    except ValueError:
        pause = 0
    if pause:
        log_debug(_('Основной поток приостановлен. Пауза %s секунд.'), pause)
        sleep(pause)
        log_debug(_('Основной поток продолжен.'))
    return pause


def _prepare_answer(answer, export_name, script, variables, source=None):
    export_name_source = export_name + '_source'
    if source is True:
        variables[export_name_source] = answer
    elif source is not None:
        variables[export_name_source] = source

    booleanize = script.get('booleanize')
    convert = script.get('convert')
    if booleanize:
        if booleanize.get('to_lowercase', True):
            answer = answer.lower()
        answer = bool(answer in booleanize['true_list'])
        text_true = booleanize.get('text_true') or _('Да')
        text_false = booleanize.get('text_false') or _('Нет')
        export_name_text = (
            booleanize.get('export_name_text') or
            export_name + '_text'
        )
        variables[export_name_text] = text_true if answer else text_false
    elif convert:
        variables[export_name + '_convert_from'] = answer
        if convert.get('to_lowercase', True):
            answer = answer.lower()
        default = convert.get('default')
        if default is None:
            default = answer
        answer = convert['values'].get(answer, default)
        labels = convert.get('labels')
        if labels and answer in labels:
            variables[export_name + '_label'] = labels[answer]
        names = convert.get('names')
        if names and answer in names:
            variables[export_name + '_name'] = labels[answer]
        texts = convert.get('texts')
        if texts and answer in texts:
            variables[export_name + '_text'] = labels[answer]
    # Устанавливаем изменённый текст ответа в главную переменную.
    variables[export_name] = answer


class Dialog:
    """Диалог робота с позвонившим на номер абонентом."""
    is_closed = False
    # Значение меняется экземпляром распознавателя в процессе считывания стрима.
    stream_size = 0
    # Значение меняется каждый раз перед вызовом парсеров.
    partials_begin = 0
    # Заполнен в момент воспроизведения голоса.
    _voice = None
    _voice_duration = 0
    _voice_nonblocking = False
    # RMS-значение данных тихого участка звукового потока используется для
    # рассчёта тишины канала и может уменьшаться в процессе работы.
    _silence_rms = 1000
    # Момент начала тишины в Unix-time (целые секунды).
    _silence_from = None
    # Флаг, при установке которого сбрасывается счётчик секунд молчания.
    _reset_silence = False

    def __init__(self, agi, settings, stream):
        self.agi = agi
        self.settings = settings
        self.stream = stream
        self.amplitude = []
        self.silence_seconds = 0

        self.synthesizer = settings.get_synthesizer()
        self.recognizer = settings.get_recognizer()
        if hasattr(self.recognizer, 'token_manager'):
            self.recognizer.token_manager.update_token()
        self.morphology_processor = settings.get_morphology_processor()
        self.text_buffers = OrderedDict()
        # Словарь для экспортируемых после завершения в Asterisk переменных.
        self.export = OrderedDict()
        # Цепочка исполнения скриптов с их результатами и распознанным текстом.
        self.chain = []

    def make_text_buffer(self, key=None):
        if not key:
            key = len(self.text_buffers)
        text_buffer = TextBuffer(key)
        self.text_buffers[text_buffer.key] = text_buffer
        log_debug(_('Новый текстовый буфер: %s'), key)
        return key

    def get_last_text_buffer(self):
        text_buffers = self.text_buffers
        return text_buffers[list(text_buffers.keys())[-1]]

    def text_to_text_buffer(self, text, fixed=False, text_buffer=None):
        if text_buffer is None:
            text_buffer = self.get_last_text_buffer()
        text = text.strip()
        if not text:
            return
        text_buffer.final = None
        # Первые передварительные результаты не принимаются до заданного
        # ограничения.
        if not fixed:
            recognizer = getattr(self, 'recognizer', None)
            if recognizer and recognizer.partials_counter < self.partials_begin:
                return
        if not text_buffer:
            text_buffer.append(text)
        else:
            text_buffer[-1] = text
        # Для блокировки делаем пустую новую строку, в неё и будет писаться
        # текст в дальнейшем.
        if fixed:
            text_buffer.fixed.append(len(text_buffer) - 1)
            text_buffer.append('')
            text_buffer.final = time()

    def text_from_text_buffer(self, mode=1, text_buffer=None):
        """
        Возвращает текст буфера в режимах:
        1 - финальный, 2 - промежуточный, 3 - весь, 4 - весь с разметкой.
        """
        if text_buffer is None:
            text_buffer = self.get_last_text_buffer()
        return text_buffer.get_text(mode)

    def parse_names(self, mode=1, text_buffer=None):
        """Метод ищет в тексте буфера все имена."""
        # Слова должны быть в оригинальной последовательности.
        text = self.text_from_text_buffer(mode, text_buffer)
        processor = self.morphology_processor
        return processor.parse_names(text), text

    def search_by_buffer(self, regexp, mode=1, text_buffer=None):
        """Метод ищет в тексте буфера по регулярному выражению."""
        text = self.text_from_text_buffer(mode, text_buffer)
        return regexp.search(text), text

    def capture_text(self, stopwords, mode=1, text_buffer=None):
        """Метод ищет в тексте буфера по регулярному выражению."""
        text = self.text_from_text_buffer(mode, text_buffer)
        stop = False
        if stopwords and stopwords.intersection(text.split(' ')):
            stop = True
        return text, stop

    def text_processing(self, name, mode=1, text_buffer=None):
        """
        Метод отправляет текст буфера в подпроцесс обработчика текста и
        возвращает результат его обработки им.
        """
        text = self.text_from_text_buffer(mode, text_buffer)
        if not text:
            return '', text
        text = text.replace('\n', ' ')
        processor = self.settings.get_processor(name)
        if getattr(self, '_text_processing_processor', '') != processor:
            self._text_processing_processor = processor
            self._text_processing_result = self._text_processing_text = ''
        # Кэшированое значение результата для текста.
        if self._text_processing_text == text:
            return self._text_processing_result, self._text_processing_text

        procname = basename(processor)
        log_debug('> PROCESSOR %s %r', procname, text)
        res = subprocess.run(
            [processor, text],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            # universal_newlines=True,
        )
        if res.returncode:
            result = ''
            log_error(_('Текстовый процессор %r сломан.'), processor)
            log_error('< PROCESSOR %s', res.stderr)
        else:
            result = res.stdout.decode('utf-8').rstrip()
            log_debug('< PROCESSOR %s %r', procname, result)
        self._text_processing_result = result
        self._text_processing_text = text
        return result, text

    def say(self, text: str, voice: str, escape_digits: str, nonblocking=False):
        if not text and not voice:
            log_error(_('Текст и файл голоса отсутствуют.'))
            return

        # Здесь "voice" - это путь к файлу без его расширения.
        if voice:
            log_info(_('Голос робота: %s'), voice)
        else:
            log_info(_('Текст робота: %s'), text)
            try:
                voice = self.synthesizer.make_voice(text)
            except Exception as e:
                log_error(_('Ошибка синтеза речи.'), exc_info=e)
                return

        n = 0
        while self._voice and n < 1000:
            log_warning(_('Другой голос %s ещё не завершён.'), self._voice)
            n += 1
            sleep(0.5)

        if self._voice:
            log_warning(_('Заменяем голос %s на новый.'), self._voice)

        self._voice = voice
        with wave.open(voice + '.wav') as f:
            duration = f.getnframes() / f.getframerate()
            log_debug(_('Продолжительность голоса %s секунд.'), duration)
            self._voice_duration = duration

        if nonblocking:
            log_debug(_('Голос отправляется в неблокирующем режиме.'))
        else:
            log_debug(_('Голос отправляется в блокирующем режиме.'))

        def clear_voice():
            if self._voice == voice:
                self._voice = None
                self._voice_duration = 0
                log_debug(_('Голос %s завершён.'), voice)
                # После завершения любой речи робота мы должны сбросить счётчик
                # молчания.
                self._reset_silence = True

        def runner():
            rest = duration
            while rest > 0 and self._voice == voice:
                sleep(0.5)
                rest -= 0.5
            clear_voice()

        try:
            self.agi.stream_file(voice, escape_digits=escape_digits)
        except AgiSigpipeError:
            self._voice = None
            self._voice_duration = 0
            self.is_closed = True
            log_info(_('Голос не отправлен. Связь разорвана.'))
            return

        thread = Thread(target=runner, daemon=True)
        thread.start()

        self._voice_nonblocking = nonblocking
        if not nonblocking:
            log_debug(_('Останавливаем основной поток на %s секунд.'), duration)
            # Останавливаем основной поток.
            sleep(duration + 0.1)
            # Запускаем гарантированную очистку.
            clear_voice()

    def say_silence(self, conf, index, variables):
        try:
            s = conf['phrases'][index]
            index += 1
        except IndexError:
            s = conf['phrases'][0]
            index = 1
        saydata = {'text': s}
        self.say(**_make_say_params(saydata, variables))
        return index

    def calc_silence_seconds(self):
        silence_rms = self._silence_rms
        # Граница тихого участка - это 200% от минимума.
        # Вроде бы смысла нет иметь границу ниже 500 единиц RMS.
        if silence_rms > 250:
            # Уменьшаем минимальное значение.
            for second, rms in self.amplitude:
                if rms < 250:
                    self._silence_rms = silence_rms = 250
                    log_debug(_('Установлен минимальный RMS=250(%d).'), rms)
                    break
                if rms < silence_rms:
                    self._silence_rms = silence_rms = rms
                    log_debug(_('Установлен минимальный RMS=%d.'), rms)

        border = silence_rms * 2
        silence_from = self._silence_from
        for second, rms in reversed(self.amplitude):
            if rms > border:
                self._silence_from = silence_from = second + 1
                break
        if silence_from is None:
            value = 0
        else:
            seconds = int(time()) - silence_from
            value = seconds if seconds > 0 else 0
        self.silence_seconds = value
        if value:
            log_debug(_('Тишина %r секунд.'), value)
        return value

    def has_start_recognize(self):
        if self.is_closed:
            return False
        agi = self.agi
        return not agi.is_sighup and not agi.is_hungup

    def start_recognizing(self):
        log_debug(_('Включаем передачу голоса на распознавание.'))
        self.recognizer.pause = False

    def stop_recognizing(self):
        log_debug(_('Отключаем передачу голоса на распознавание.'))
        self.recognizer.pause = True

    @property
    def is_stopped(self):
        if self.is_closed:
            return True
        agi = self.agi
        return agi.is_sighup or agi.is_hungup

    def start(self):
        if getattr(self, '_recognizer_thread', None):
            raise RuntimeError(_('Другой поток распознавания уже запущен.'))
        if not self.text_buffers:
            raise RuntimeError(_('Начальный текстовый буфер удалён.'))

        def calc_silence():
            # От начала и до конца беседы поток голоса абонента анализируется
            # на тишину каждую секунду.
            while self.has_start_recognize():
                if self._reset_silence:
                    # self.amplitude.clear()
                    self._silence_from = int(time())
                    self._reset_silence = False
                    log_debug(_('Сброс счётчика молчания.'))
                sleep(1)
                self.calc_silence_seconds()
            thread = self._calc_silence_thread
            self._calc_silence_thread = None
            log_debug(_('Завершён поток подсчёта тишины %s.'), thread)

        log_debug(_('Запуск отдельного потока подсчёта тишины.'))
        thread = Thread(target=calc_silence, daemon=True)
        thread.start()
        log_debug(_('Запущен поток подсчёта тишины %s.'), thread)
        self._calc_silence_thread = thread

        def listen():
            # От начала и до конца беседы поток голоса абонента анализируется.
            # Причём, когда объём потока превышает разрешённые 10 мегабайт
            # и парсинг завершается, мы запускаем новый парсинг.
            while self.has_start_recognize():
                log_debug(_('Запуск recognizer.recognize()'))
                self.recognizer.listen(self)
                log_debug(_('Остановка recognizer.recognize()'))
            thread = self._recognizer_thread
            self._recognizer_thread = None
            log_debug(_('Завершён поток распознавания %s.'), thread)

        self.is_closed = False
        log_debug(_('Запуск отдельного потока распознавания.'))
        thread = Thread(target=listen, daemon=True)
        thread.start()
        log_debug(_('Запущен поток распознавания %s.'), thread)
        self._recognizer_thread = thread
        return thread

    def stop(self):
        log_debug(_('Диалог останавливается.'))
        self.is_closed = True
        # Отключаем здесь листенер, чтобы быстрее выйти, а не ждать слова в
        # трубку.
        if hasattr(self.recognizer, 'listener'):
            listener = self.recognizer.listener
            listener.stop()
            log_debug(_('Слушатель выключен.'))
        else:
            log_error(_('Слушатель не был подключен.'))

    def run(self):
        log_info(
            _('Диалог начат: %(agi_callerid)s (%(agi_calleridname)s)'),
            self.agi.params,
        )
        # Делаем самый первый текстовый буфер.
        self.make_text_buffer()
        # Затем запускаем процесс потокового распознавания речи.
        self.start()
        hangup = False
        error = None
        started = datetime.utcnow()
        try:
            self.main_script()
        except AgiHangupError:
            log_info(_('Абонент повесил трубку до окончания диалога.'))
            hangup = True
        except AgiError as e:
            error = _('Ошибка в AGI при исполнении диалога.')
            log_error(error, exc_info=e)
        except Exception as e:
            error = _('Ошибка в сценарии диалога.')
            log_error(error, exc_info=e)
        self.stop()
        finished = datetime.utcnow()

        agi = self.agi
        try:
            for key, value in self.export.items():
                if key.startswith('_'):
                    continue
                value = str(value)
                log_debug(_('Экспорт %(key)s=%(value)s'), {
                    'key': key, 'value': value})
                agi.set(key, value)
        except AgiHangupError:
            log_info(
                _('Абонент повесил трубку до окончания передачи переменных.'))
            hangup = True
        except AgiError as e:
            error = _('Ошибка в AGI при передаче переменных.')
            log_error(error, exc_info=e)

        filename = self.settings.export_dialog(
            self.agi.params, self.export, self.text_buffers,
            hangup=hangup,
            error=error,
            chain=self.chain,
            started='%sZ' % started.isoformat(),
            finished='%sZ' % finished.isoformat(),
            duration=(finished - started).total_seconds(),
        )
        if filename:
            log_info(_('Диалог экспортирован в файл %s'), filename)

        for key, text_buffer in self.text_buffers.items():
            for line in text_buffer.get_text(4).split('\n'):
                log_debug(_('Текст буфера %(key)s: %(line)s'),
                          {'key': key, 'line': line})

        log_info(_('Диалог завершён.'))

    def execute_plugin(self, name, result, speech, variables):
        """
        Возвращает кортеж из результата выполнения подпроцесса и словаря данных,
        которыми плагин оперировал.
        Плагин может устанавливать свои переменные в процессе своей работы или
        ещё какие-либо данные в файл обмена данными.
        """
        plugin = self.settings.get_plugin(name)
        log_debug('> PLUGIN %s', plugin)
        plugin_data = {
            'agi_params': self.agi.params,
            'work_dir': self.settings.work_dir,
            'speech': speech,
            'variables': variables,
            'chain': self.chain,
            'text_buffers': self.text_buffers,
        }
        if result is None or \
                isinstance(result, (str, bool, list, tuple, int, float)):
            plugin_data['result'] = result
        else:
            plugin_data['result'] = str(result)
        tempfile = NamedTemporaryFile(mode='w', suffix='.json')
        tempfile_name = tempfile.name
        json.dump(plugin_data, tempfile)
        tempfile.flush()
        log_debug('JSONFILE %s', tempfile_name)
        res = subprocess.run(
            [plugin, tempfile_name],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

        if res.returncode:
            log_error(_('Плагин %r сломан.'), plugin)
            log_error('< PLUGIN %s', res.stderr)
        else:
            log_debug('< PLUGIN %s', res.stdout)
            plugin_data = json.load(open(tempfile_name))
            variables.update(plugin_data['variables'])
            # Предварительный фоновый синтез текстов.
            synthesize = plugin_data.get('synthesize', [])
            if synthesize and isinstance(synthesize, list):
                def make_voices():
                    log_debug(
                        _('Предварительный синтез для %d текста(ов).'),
                        len(synthesize)
                    )
                    for txt in synthesize:
                        self.synthesizer.make_voice(str(txt), early=True)
                    log_debug(_('Предварительный синтез речи завершён.'))
                t = Thread(target=make_voices, daemon=True)
                t.start()
            # Логгирование каких-либо строк из плагина.
            loggers = (
                ('debug', log_debug),
                ('info', log_info),
                ('warnings', log_warning),
                ('errors', log_error),
            )
            for _name, _logger in loggers:
                strings = plugin_data.get(_name, [])
                for string in strings:
                    _logger('< PLUGIN %s', string)
        tempfile.close()
        return res, plugin_data

    def main_script(self):
        scheme = self.settings.dialog_scheme
        constants = scheme.get('constants', {})
        self.export = variables = OrderedDict()
        self.chain = chain = []
        variables.update(constants)
        # Словарь скриптов.
        scripts = scheme.get('scripts', {})
        # Программа главного скрипта.
        program = OrderedDict([(p['name'], p) for p in scheme['main_script']])
        index = tuple(program.keys())
        process = program[index[0]]
        log_debug(_('Главный скрипт начат.'))
        while process and not self.is_closed:
            name = process['name']
            script = scripts.get(name)
            if not script:
                break
            scrypt_type = script['type']
            # Здесь должна возникать ошибка AttributeError для сломанных схем.
            method = getattr(self, '_script_%s' % scrypt_type)
            log_debug('  SCRIPT %s', name.upper())
            self.partials_begin = 0
            if 'partials_begin' in script:
                try:
                    self.partials_begin = int(script['partials_begin'])
                except ValueError:
                    pass

            say_debug = process.get('say_debug')
            if say_debug:
                saydata = {'text': _('Запускаю скрипт %(name)r') % process}
                self.say(**_make_say_params(saydata, variables))

            result, speech = method(script, variables)
            log_debug('  RESULT %r', result)
            log_debug('  SPEECH %r', speech)
            # Экспортируем разговор в отдельную переменную, чтобы иметь
            # доступ к нему в других скриптах и плагинах.
            export_speech = script.get('export_speech')
            if export_speech:
                variables[export_speech] = speech

            # Запускаем плагин при его наличии.
            plugin_name = process.get('plugin')
            if plugin_name:
                self.execute_plugin(plugin_name, result, speech, variables)

            # Устанавливаем следующий процесс.
            success = process.get('success')
            if isinstance(success, dict):
                success = success.get(result)
            # Динамическое определение следующего скрипта может определяться
            # из ранее установленных переменных.
            if success and '%(' in success:
                success %= variables
            fail = process.get('fail')
            if fail and '%(' in fail:
                fail %= variables
            log_debug(' SUCCESS %s', success)
            log_debug('    FAIL %s', fail)
            if process.get('finish', False):
                log_debug('  FINISH')
                if say_debug:
                    saydata = {'text': _('Завершаю диалог.')}
                    self.say(**_make_say_params(saydata, variables))
                process = None
            elif result and success in program:
                if say_debug:
                    saydata = {'text': _('Исходная речь. %s.') % speech}
                    self.say(**_make_say_params(saydata, variables))
                    if result is True:
                        saydata = {'text': _('В результате получена истина.')}
                    else:
                        saydata = {
                            'text': _('Разультат обработки. %s.') % result}
                    self.say(**_make_say_params(saydata, variables))
                process = program[success]
            elif not result and fail in program:
                if say_debug:
                    if speech:
                        saydata = {'text': _('Исходная речь. %s.') % speech}
                    else:
                        saydata = {'text': _('Исходная речь пуста.')}
                    self.say(**_make_say_params(saydata, variables))
                    saydata = {'text': _('Результат отсутствует.')}
                    self.say(**_make_say_params(saydata, variables))
                process = program[fail]
            else:
                try:
                    process = program.get(index[index.index(name) + 1])
                except IndexError:
                    process = None
            log_debug('    NEXT %s', process)
            # if say_debug and process:
            #     saydata = {'text': _('Следующий скрипт %(name)r') % process}
            #     self.say(**_make_say_params(saydata, variables))

            # Восстанавливаем константы. Это проще, чем городить лес кода.
            variables.update(constants)
            # Сохраняем цепочку вызовов скриптов.
            chain.append([name, result, speech])

        log_debug(_('Главный скрипт завершён.'))

    def _agi_execute(self, commands, key):
        try:
            command = commands.get(key, [])
            assert isinstance(command, list)
        except AssertionError:
            command = []
        if command:
            # Обрабатываем аргументы, преобразовывая шаблонные переменные.
            kw = self.agi.params.copy()
            kw.update(self.export)
            for i, val in enumerate(command):
                # Пропускаем первый аргумент и все не строковые аргументы.
                if i and isinstance(val, str):
                    command[i] = val.format(**kw)
            log_debug(_('Выполняется команда AGI: %s'), command)
            self.agi.execute(*command)
            log_debug(_('Команда выполнена.'))
        return command

    def _agi_stop_playback(self, saydata=None):
        log_debug('_agi_stop_playback')
        if not saydata or not isinstance(saydata, dict):
            log_debug(_('Нечего опрерывать, saydata=%s'), saydata)
            return
        if not saydata.get('nonblocking', False):
            log_debug(_('Нечего прерывать, был задан блокирующий режим.'))
            return

        stop_voice = splitext(self.settings.get_stop_file())[0]
        if not stop_voice:
            log_debug(_('Стоп-голоса для прерывания нет.'))
            return
        if not self._voice:
            log_debug(_('Нечего прерывать, голос уже закончился.'))
            return
        # Сначала должен среагировать поток озвучки голоса, а затем отправляем
        # команду остановки.
        self._voice = None
        self._voice_duration = 0
        try:
            self.agi.stream_file(stop_voice)
        except AgiSigpipeError:
            self.is_closed = True
            log_info(_('Стоп-голос не отправлен. Связь разорвана.'))
            return
        except Exception as e:
            self.is_closed = True
            raise e
        log_debug(_('Прерывание воспроизведения выполнено.'))

    def _script_capture(self, script, variables):
        assert script['type'] == 'capture'

        assert 'export_name' in script
        export_name = script['export_name']
        assert export_name and isinstance(export_name, str)

        pauses = script.get('pauses', {})
        _make_pause(pauses, 'before')

        commands = script.get('commands', {})
        self._agi_execute(commands, 'before')

        mode = int(script.get('mode_buffer', 1))
        assert mode in (1, 2, 3)
        # Для капчуринга всегда создаём новый буфер.
        self.make_text_buffer()
        text_buffer = self.get_last_text_buffer()

        saydata = script.get('say_start')
        if saydata:
            _make_pause(pauses, 'before_say_start')
            self._agi_execute(commands, 'before_say_start')
            self.say(**_make_say_params(saydata, variables))
            _make_pause(pauses, 'after_say_start')
            self._agi_execute(commands, 'after_say_start')
        say_start_duration = self._voice_duration

        self.start_recognizing()

        maxtime = float(script.get('maxtime', 600))
        minlength = int(script.get('minlength', 10))
        silence = int(script.get('silence', 15))
        stopwords = script.get('stopwords', [])
        log_debug(_('Ограничение по времени: %s секунд.'), maxtime)
        log_debug(_('Ограничение по длине: от %s символов.'), minlength)
        log_debug(_('Ограничение по тишине: %s секунд.'), stopwords)
        log_debug(_('Ограничение по стоп-словам: %s.'), stopwords)
        stopwords = set(stopwords)

        func = self.capture_text
        speech, stop = func(stopwords, mode=mode)
        if stop:
            log_debug(_('Первая попытка успешна. Текст: %s'), speech)
        else:
            log_debug(_('Первая попытка провалилась. Текст: %s'), speech)
        # В блокирующем режиме голоса его длина уже будет сброшена в 0.
        maxtime += say_start_duration

        second = 0
        # Количество секунд тишины.
        while not stop and second < maxtime:
            log_debug(_('Стоп-слова пока нет, ждём %d-ю секунду.'), second)
            second += 0.5
            sleep(0.5)
            speech, stop = func(stopwords, mode=mode)
            if stop:
                log_debug('Сработало ограничение по стоп-слову.')
            else:
                silence_seconds = self.silence_seconds
                _expire = silence_seconds >= silence
                _size = len(speech) >= minlength
                _final = text_buffer.seconds_after_final > 1
                stop = _expire and _size and _final
                if _expire:
                    log_debug('Сработало ограничение по тишине.')
                    if _size:
                        log_debug('Сработало ограничение по длине.')
                        if _final:
                            log_debug('Сработало ограничение по финалу '
                                      'распознавания.')
                        else:
                            log_debug('Ожидаем финальное распознавание.')
                    else:
                        log_debug('Длина текста недостаточна для завершения.')

            if self.is_stopped:
                log_debug(_('Диалог завершился до приёма данных.'))
                break

        self.stop_recognizing()

        if speech:
            answer = speech
            log_debug(_('Ответ: %r'), answer)
            _prepare_answer(
                answer=answer, source=answer, export_name=export_name,
                script=script, variables=variables)

            self._agi_stop_playback(script.get('say_start'))

            # Когда указан файл плагина, то он должен обработать переменные
            # перед say_success.
            plugin_name = script.get('plugin')
            if plugin_name:
                self.execute_plugin(plugin_name, answer, speech, variables)

            saydata = script.get('say_success')
            if saydata:
                _make_pause(pauses, 'before_say_success')
                self._agi_execute(commands, 'before_say_success')
                self.say(**_make_say_params(saydata, variables))
                _make_pause(pauses, 'after_say_success')
                self._agi_execute(commands, 'after_say_success')
        else:
            answer = None
            log_debug(_('Ответа нет.'))
            saydata = script.get('say_fail')
            if saydata:
                _make_pause(pauses, 'before_say_fail')
                self._agi_execute(commands, 'before_say_fail')
                self.say(**_make_say_params(saydata, variables))
                _make_pause(pauses, 'after_say_fail')
                self._agi_execute(commands, 'after_say_fail')

        _make_pause(pauses, 'after')
        self._agi_execute(commands, 'after')
        return answer, speech

    def _script_parse_names(self, script, variables):
        assert script['type'] == 'parse_names'

        assert 'export_name' in script
        export_name = script['export_name']
        assert export_name and isinstance(export_name, str)

        pauses = script.get('pauses', {})
        _make_pause(pauses, 'before')

        commands = script.get('commands', {})
        self._agi_execute(commands, 'before')

        mode = int(script.get('mode_buffer', 1))
        assert mode in (1, 2, 3)
        if script.get('new_buffer', False):
            self.make_text_buffer()

        saydata = script.get('say_start')
        if saydata:
            _make_pause(pauses, 'before_say_start')
            self._agi_execute(commands, 'before_say_start')
            self.say(**_make_say_params(saydata, variables))
            _make_pause(pauses, 'after_say_start')
            self._agi_execute(commands, 'after_say_start')
        say_start_duration = self._voice_duration

        self.start_recognizing()

        say_silence = script.get('say_silence')
        if say_silence:
            say_silence_time = say_silence['time']
        else:
            say_silence_time = 0
        say_silence_index = 0

        attempts = script.get('attempts') or {}
        attempts_count = abs(int(attempts.get('count', 0))) or 1
        attempts_iter_time = float(attempts.get('iter_time', 0))
        attempts_iter_pause = float(attempts.get('iter_pause', 0)) or 1.0

        func = self.parse_names
        names, speech = func(mode=mode)
        if names:
            log_debug(_('Первая попытка успешна. Имена: %s'), names)
        else:
            log_debug(_('Первая попытка провалилась. Имена: %s'), names)
        for attempt in range(attempts_count):
            _iter_time = attempts_iter_time
            if attempt == 0:
                # В блокирующем режиме голоса его длина уже будет сброшена в 0.
                _iter_time += say_start_duration
            second = 0
            while not names and second < _iter_time:
                log_debug(_('Ответа пока нет, ждём %d-ю секунду.'), second)
                second += attempts_iter_pause
                sleep(attempts_iter_pause)
                names, speech = func(mode=mode)
                if self.is_stopped:
                    log_debug(_('Диалог завершился до приёма данных.'))
                    break
                # Обработка молчания ягнят.
                if say_silence_time and (attempt > 0 or second > say_start_duration):
                    ss = self.silence_seconds
                    if second and ss and ss % say_silence_time == 0:
                        log_debug(_('Кратность %r.'), say_silence_time)
                        say_silence_index = self.say_silence(
                            say_silence, say_silence_index, variables)
            if names:
                length_names = int(script.get('length_names', 2))
                answer = ' '.join(names[:length_names])
                log_debug(_('Ответ: %r'), answer)
                _prepare_answer(
                    answer=answer, source=answer, export_name=export_name,
                    script=script, variables=variables)

                self._agi_stop_playback(script.get('say_start'))

                saydata = script.get('say_success')
                if saydata:
                    _make_pause(pauses, 'before_say_success')
                    self._agi_execute(commands, 'before_say_success')
                    self.say(**_make_say_params(saydata, variables))
                    _make_pause(pauses, 'after_say_success')
                    self._agi_execute(commands, 'after_say_success')
                break
            elif attempt < attempts_count - 1:
                if attempts.get('new_buffer', False):
                    self.make_text_buffer()
                saydata = attempts.get('say')
                if saydata:
                    _make_pause(pauses, 'before_say_attempt')
                    self._agi_execute(commands, 'before_say_attempt')
                    self.say(**_make_say_params(saydata, variables))
                    _make_pause(pauses, 'after_say_attempt')
                    self._agi_execute(commands, 'after_say_attempt')

        self.stop_recognizing()

        if not names:
            log_debug(_('Ответа нет.'))
            saydata = script.get('say_fail')
            if saydata:
                _make_pause(pauses, 'before_say_fail')
                self._agi_execute(commands, 'before_say_fail')
                self.say(**_make_say_params(saydata, variables))
                _make_pause(pauses, 'after_say_fail')
                self._agi_execute(commands, 'after_say_fail')

        _make_pause(pauses, 'after')
        self._agi_execute(commands, 'after')
        return names, speech

    def _script_search_by_buffer(self, script, variables):
        assert script['type'] == 'search_by_buffer'

        assert 'export_name' in script
        export_name = script['export_name']
        assert export_name and isinstance(export_name, str)

        assert 'regexp' in script
        regexp = script['regexp']
        assert regexp and isinstance(regexp, dict)

        assert 'pattern' in regexp
        pattern = regexp['pattern']
        assert pattern and isinstance(pattern, str)

        pauses = script.get('pauses', {})
        _make_pause(pauses, 'before')

        commands = script.get('commands', {})
        self._agi_execute(commands, 'before')

        mode = int(script.get('mode_buffer', 1))
        assert mode in (1, 2, 3)
        if script.get('new_buffer', False):
            self.make_text_buffer()

        saydata = script.get('say_start')
        if saydata:
            _make_pause(pauses, 'before_say_start')
            self._agi_execute(commands, 'before_say_start')
            self.say(**_make_say_params(saydata, variables))
            _make_pause(pauses, 'after_say_start')
            self._agi_execute(commands, 'after_say_start')
        say_start_duration = self._voice_duration

        self.start_recognizing()

        say_silence = script.get('say_silence')
        if say_silence:
            say_silence_time = say_silence['time']
        else:
            say_silence_time = 0
        say_silence_index = 0

        attempts = script.get('attempts') or {}
        attempts_count = abs(int(attempts.get('count', 0))) or 1
        attempts_iter_time = float(attempts.get('iter_time', 0))
        attempts_iter_pause = float(attempts.get('iter_pause', 0)) or 1.0

        func = self.search_by_buffer
        flag = re.UNICODE
        if regexp.get('ignorecase', False):
            flag |= re.IGNORECASE
        pattern = re.compile(pattern, flag)
        match, speech = func(pattern, mode=mode)
        answer = None
        for attempt in range(attempts_count):
            _iter_time = attempts_iter_time
            if attempt == 0:
                # В блокирующем режиме голоса его длина уже будет сброшена в 0.
                _iter_time += say_start_duration
            second = 0
            while not match and second < _iter_time:
                log_debug(_('Ответа пока нет, ждём %d-ю секунду.'), second)
                second += attempts_iter_pause
                sleep(attempts_iter_pause)
                match, speech = func(pattern, mode=mode)
                if self.is_stopped:
                    log_debug(_('Диалог завершился до приёма данных.'))
                    break
                # Обработка молчания ягнят.
                if say_silence_time and (attempt > 0 or second > say_start_duration):
                    ss = self.silence_seconds
                    if second and ss and ss % say_silence_time == 0:
                        log_debug(_('Кратность %r.'), say_silence_time)
                        say_silence_index = self.say_silence(
                            say_silence, say_silence_index, variables)
            if match:
                answer = match.group()
                log_debug(_('Ответ: %r'), answer)
                _prepare_answer(
                    answer=answer, source=answer, export_name=export_name,
                    script=script, variables=variables)

                self._agi_stop_playback(script.get('say_start'))

                saydata = script.get('say_success')
                if saydata:
                    _make_pause(pauses, 'before_say_success')
                    self._agi_execute(commands, 'before_say_success')
                    self.say(**_make_say_params(saydata, variables))
                    _make_pause(pauses, 'after_say_success')
                    self._agi_execute(commands, 'after_say_success')
                break
            elif attempt < attempts_count - 1:
                if attempts.get('new_buffer', False):
                    self.make_text_buffer()
                saydata = attempts.get('say')
                if saydata:
                    _make_pause(pauses, 'before_say_attempt')
                    self._agi_execute(commands, 'before_say_attempt')
                    self.say(**_make_say_params(saydata, variables))
                    _make_pause(pauses, 'after_say_attempt')
                    self._agi_execute(commands, 'after_say_attempt')

        self.stop_recognizing()

        if not match:
            log_debug(_('Ответа нет.'))
            saydata = script.get('say_fail')
            if saydata:
                _make_pause(pauses, 'before_say_fail')
                self._agi_execute(commands, 'before_say_fail')
                self.say(**_make_say_params(saydata, variables))
                _make_pause(pauses, 'after_say_fail')
                self._agi_execute(commands, 'after_say_fail')

        _make_pause(pauses, 'after')
        self._agi_execute(commands, 'after')
        if script.get('fail_on_false'):
            answer = variables.get(export_name) in (
                True, 'True', 'true', 'TRUE')
            log_debug(_('Режим `fail_on_false`, ответ %r.'), answer)
        return answer, speech

    def _script_text_processing(self, script, variables):
        assert script['type'] == 'text_processing'

        assert 'export_name' in script
        export_name = script['export_name']
        assert export_name and isinstance(export_name, str)

        assert 'file' in script
        filename = script['file']
        assert filename and isinstance(filename, str)

        pauses = script.get('pauses', {})
        _make_pause(pauses, 'before')

        commands = script.get('commands', {})
        self._agi_execute(commands, 'before')

        mode = int(script.get('mode_buffer', 1))
        assert mode in (1, 2, 3)
        if script.get('new_buffer', False):
            self.make_text_buffer()

        saydata = script.get('say_start')
        if saydata:
            _make_pause(pauses, 'before_say_start')
            self._agi_execute(commands, 'before_say_start')
            self.say(**_make_say_params(saydata, variables))
            _make_pause(pauses, 'after_say_start')
            self._agi_execute(commands, 'after_say_start')
        say_start_duration = self._voice_duration

        self.start_recognizing()

        say_silence = script.get('say_silence')
        if say_silence:
            say_silence_time = say_silence['time']
        else:
            say_silence_time = 0
        say_silence_index = 0

        attempts = script.get('attempts') or {}
        attempts_count = abs(int(attempts.get('count', 0))) or 1
        attempts_iter_time = float(attempts.get('iter_time', 0))
        attempts_iter_pause = float(attempts.get('iter_pause', 0)) or 1.0

        func = self.text_processing
        response, speech = func(filename, mode=mode)
        if response:
            log_debug(_('Первая попытка успешна. Ответ: %s'), response)
        else:
            log_debug(_('Первая попытка провалилась. Ответа нет.'))
        for attempt in range(attempts_count):
            _iter_time = attempts_iter_time
            if attempt == 0:
                # В блокирующем режиме голоса его длина уже будет сброшена в 0.
                _iter_time += say_start_duration
            second = 0
            while not response and second < _iter_time:
                log_debug(_('Ответа пока нет, ждём %d-ю секунду.'), second)
                second += attempts_iter_pause
                sleep(attempts_iter_pause)
                response, speech = func(filename, mode=mode)
                if self.is_stopped:
                    log_debug(_('Диалог завершился до приёма данных.'))
                    break
                # Обработка молчания ягнят.
                if say_silence_time and (attempt > 0 or second > say_start_duration):
                    ss = self.silence_seconds
                    if second and ss and ss % say_silence_time == 0:
                        log_debug(_('Кратность %r.'), say_silence_time)
                        say_silence_index = self.say_silence(
                            say_silence, say_silence_index, variables)
            if response:
                answer = response
                log_debug(_('Ответ: %r'), answer)
                _prepare_answer(
                    answer=answer, source=speech, export_name=export_name,
                    script=script, variables=variables)

                self._agi_stop_playback(script.get('say_start'))

                saydata = script.get('say_success')
                if saydata:
                    _make_pause(pauses, 'before_say_success')
                    self._agi_execute(commands, 'before_say_success')
                    self.say(**_make_say_params(saydata, variables))
                    _make_pause(pauses, 'after_say_success')
                    self._agi_execute(commands, 'after_say_success')
                break
            elif attempt < attempts_count - 1:
                if attempts.get('new_buffer', False):
                    self.make_text_buffer()
                saydata = attempts.get('say')
                if saydata:
                    _make_pause(pauses, 'before_say_attempt')
                    self._agi_execute(commands, 'before_say_attempt')
                    self.say(**_make_say_params(saydata, variables))
                    _make_pause(pauses, 'after_say_attempt')
                    self._agi_execute(commands, 'after_say_attempt')

        self.stop_recognizing()

        if not response:
            log_debug(_('Ответа нет.'))
            saydata = script.get('say_fail')
            if saydata:
                _make_pause(pauses, 'before_say_fail')
                self._agi_execute(commands, 'before_say_fail')
                self.say(**_make_say_params(saydata, variables))
                _make_pause(pauses, 'after_say_fail')
                self._agi_execute(commands, 'after_say_fail')

        _make_pause(pauses, 'after')
        self._agi_execute(commands, 'after')
        return response, speech

    def _script_simple_say(self, script, variables):
        assert script['type'] == 'simple_say'

        if 'export_name' in script and 'export_value' in script:
            variables[script['export_name']] = str(script['export_value'])

        pauses = script.get('pauses', {})
        _make_pause(pauses, 'before')

        commands = script.get('commands', {})
        self._agi_execute(commands, 'before')

        mode = int(script.get('mode_buffer', 1))
        assert mode in (1, 2)
        if script.get('new_buffer', False):
            self.make_text_buffer()

        plugin_name = script.get('plugin_before')
        if plugin_name:
            res, plugin_data = self.execute_plugin(plugin_name, '', '', variables)
            result = plugin_data.get('result')
        else:
            result = True

        saydata = script.get('say_start')
        if saydata:
            _make_pause(pauses, 'before_say_start')
            self._agi_execute(commands, 'before_say_start')
            self.say(**_make_say_params(saydata, variables))
            _make_pause(pauses, 'after_say_start')
            self._agi_execute(commands, 'after_say_start')

        plugin_name = script.get('plugin_after')
        if plugin_name:
            res, plugin_data = self.execute_plugin(plugin_name, result, '', variables)
            result = plugin_data.get('result')

        _make_pause(pauses, 'after')
        self._agi_execute(commands, 'after')

        speech = self.text_from_text_buffer()
        return result, speech



#
# Copyright (c) 2020, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
from audioop import rms
from logging import getLogger
from time import time


class Listener:
    chunk_size = 4000
    # Максимальный размер потокового файла для Яндекса - 10 мегабайт.
    max_size = 10485760
    amplitude_size = 2
    sample_width = 2
    total_size = 0
    _continue = True
    logger = getLogger('astersay.backends.Listener')

    def __init__(self, stream, amplitude=None):
        """
        :param stream: файлоподобный обЪект потока голоса.
        :param amplitude: список (пустой) для записи результатов посекундного
                          анализа потока.
        :type amplitude: list
        """
        self._stream = stream
        if amplitude is None:
            amplitude = []
        self.amplitude = amplitude

    def has_continue(self):
        return self._continue and (
            self.total_size + self.chunk_size < self.max_size)

    def listen(self):
        chunk_size = self.chunk_size
        stream_read = self._stream.read
        logger_debug = self.logger.debug

        amplitude = self.amplitude
        amplitude_size = self.amplitude_size
        sample_width = self.sample_width
        chunk_number = 0
        sample = b''
        sample_second = int(time())
        while self.has_continue():
            chunk_number += 1
            chunk_second = int(time())
            chunk = stream_read(chunk_size)
            length = len(chunk)
            if sample_second < chunk_second:
                rms_value = rms(sample, sample_width)
                # logger_debug('rms=%d', rms_value)
                amplitude.append((sample_second, rms_value))
                if len(amplitude) > amplitude_size:
                    amplitude.pop(0)
                sample = b''
                sample_second = chunk_second
            sample += chunk
            # logger_debug('chunck %d length is %d', chunk_number, length)
            if length == 0:
                self.stop()
            self.total_size += length
            yield chunk
        logger_debug('listener ended, total size %d', self.total_size)

    def stop(self):
        self._continue = False
        self.logger.debug('listener stopped')


class BaseBackend:

    def __init__(self):
        cls = self.__class__
        name = '%s.%s' % (cls.__module__, cls.__name__)
        self.logger = getLogger(name)

    def __str__(self):
        return str(self.__class__)[8:-2]



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: stt_service.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
  name='stt_service.proto',
  package='yandex.cloud.ai.stt.v2',
  syntax='proto3',
  serialized_options=_b('Z>github.com/yandex-cloud/go-genproto/yandex/cloud/ai/stt/v2;stt'),
  serialized_pb=_b('\n\x11stt_service.proto\x12\x16yandex.cloud.ai.stt.v2\"\x88\x01\n\x1bStreamingRecognitionRequest\x12;\n\x06\x63onfig\x18\x01 \x01(\x0b\x32).yandex.cloud.ai.stt.v2.RecognitionConfigH\x00\x12\x17\n\raudio_content\x18\x02 \x01(\x0cH\x00\x42\x13\n\x11streaming_request\"f\n\x11RecognitionConfig\x12>\n\rspecification\x18\x01 \x01(\x0b\x32\'.yandex.cloud.ai.stt.v2.RecognitionSpec\x12\x11\n\tfolder_id\x18\x02 \x01(\t\"\xbf\x02\n\x0fRecognitionSpec\x12M\n\x0e\x61udio_encoding\x18\x01 \x01(\x0e\x32\x35.yandex.cloud.ai.stt.v2.RecognitionSpec.AudioEncoding\x12\x19\n\x11sample_rate_hertz\x18\x02 \x01(\x03\x12\x15\n\rlanguage_code\x18\x03 \x01(\t\x12\x18\n\x10profanity_filter\x18\x04 \x01(\x08\x12\r\n\x05model\x18\x05 \x01(\t\x12\x17\n\x0fpartial_results\x18\x07 \x01(\x08\x12\x18\n\x10single_utterance\x18\x08 \x01(\x08\"O\n\rAudioEncoding\x12\x1e\n\x1a\x41UDIO_ENCODING_UNSPECIFIED\x10\x00\x12\x10\n\x0cLINEAR16_PCM\x10\x01\x12\x0c\n\x08OGG_OPUS\x10\x02\"\x7f\n\x1cStreamingRecognitionResponse\x12>\n\x06\x63hunks\x18\x01 \x03(\x0b\x32..yandex.cloud.ai.stt.v2.SpeechRecognitionChunk\x12\x1f\n\x17\x65nd_of_single_utterance\x18\x02 \x01(\x08\"s\n\x16SpeechRecognitionChunk\x12J\n\x0c\x61lternatives\x18\x01 \x03(\x0b\x32\x34.yandex.cloud.ai.stt.v2.SpeechRecognitionAlternative\x12\r\n\x05\x66inal\x18\x02 \x01(\x08\"@\n\x1cSpeechRecognitionAlternative\x12\x0c\n\x04text\x18\x01 \x01(\t\x12\x12\n\nconfidence\x18\x02 \x01(\x02\x32\x94\x01\n\nSttService\x12\x85\x01\n\x12StreamingRecognize\x12\x33.yandex.cloud.ai.stt.v2.StreamingRecognitionRequest\x1a\x34.yandex.cloud.ai.stt.v2.StreamingRecognitionResponse\"\x00(\x01\x30\x01\x42@Z>github.com/yandex-cloud/go-genproto/yandex/cloud/ai/stt/v2;sttb\x06proto3')
)



_RECOGNITIONSPEC_AUDIOENCODING = _descriptor.EnumDescriptor(
  name='AudioEncoding',
  full_name='yandex.cloud.ai.stt.v2.RecognitionSpec.AudioEncoding',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='AUDIO_ENCODING_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LINEAR16_PCM', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OGG_OPUS', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=529,
  serialized_end=608,
)
_sym_db.RegisterEnumDescriptor(_RECOGNITIONSPEC_AUDIOENCODING)


_STREAMINGRECOGNITIONREQUEST = _descriptor.Descriptor(
  name='StreamingRecognitionRequest',
  full_name='yandex.cloud.ai.stt.v2.StreamingRecognitionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='config', full_name='yandex.cloud.ai.stt.v2.StreamingRecognitionRequest.config', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='audio_content', full_name='yandex.cloud.ai.stt.v2.StreamingRecognitionRequest.audio_content', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='streaming_request', full_name='yandex.cloud.ai.stt.v2.StreamingRecognitionRequest.streaming_request',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=46,
  serialized_end=182,
)


_RECOGNITIONCONFIG = _descriptor.Descriptor(
  name='RecognitionConfig',
  full_name='yandex.cloud.ai.stt.v2.RecognitionConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='specification', full_name='yandex.cloud.ai.stt.v2.RecognitionConfig.specification', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='folder_id', full_name='yandex.cloud.ai.stt.v2.RecognitionConfig.folder_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=184,
  serialized_end=286,
)


_RECOGNITIONSPEC = _descriptor.Descriptor(
  name='RecognitionSpec',
  full_name='yandex.cloud.ai.stt.v2.RecognitionSpec',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='audio_encoding', full_name='yandex.cloud.ai.stt.v2.RecognitionSpec.audio_encoding', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sample_rate_hertz', full_name='yandex.cloud.ai.stt.v2.RecognitionSpec.sample_rate_hertz', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='language_code', full_name='yandex.cloud.ai.stt.v2.RecognitionSpec.language_code', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='profanity_filter', full_name='yandex.cloud.ai.stt.v2.RecognitionSpec.profanity_filter', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='model', full_name='yandex.cloud.ai.stt.v2.RecognitionSpec.model', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='partial_results', full_name='yandex.cloud.ai.stt.v2.RecognitionSpec.partial_results', index=5,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='single_utterance', full_name='yandex.cloud.ai.stt.v2.RecognitionSpec.single_utterance', index=6,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _RECOGNITIONSPEC_AUDIOENCODING,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=289,
  serialized_end=608,
)


_STREAMINGRECOGNITIONRESPONSE = _descriptor.Descriptor(
  name='StreamingRecognitionResponse',
  full_name='yandex.cloud.ai.stt.v2.StreamingRecognitionResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='chunks', full_name='yandex.cloud.ai.stt.v2.StreamingRecognitionResponse.chunks', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='end_of_single_utterance', full_name='yandex.cloud.ai.stt.v2.StreamingRecognitionResponse.end_of_single_utterance', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=610,
  serialized_end=737,
)


_SPEECHRECOGNITIONCHUNK = _descriptor.Descriptor(
  name='SpeechRecognitionChunk',
  full_name='yandex.cloud.ai.stt.v2.SpeechRecognitionChunk',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='alternatives', full_name='yandex.cloud.ai.stt.v2.SpeechRecognitionChunk.alternatives', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='final', full_name='yandex.cloud.ai.stt.v2.SpeechRecognitionChunk.final', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=739,
  serialized_end=854,
)


_SPEECHRECOGNITIONALTERNATIVE = _descriptor.Descriptor(
  name='SpeechRecognitionAlternative',
  full_name='yandex.cloud.ai.stt.v2.SpeechRecognitionAlternative',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='text', full_name='yandex.cloud.ai.stt.v2.SpeechRecognitionAlternative.text', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='confidence', full_name='yandex.cloud.ai.stt.v2.SpeechRecognitionAlternative.confidence', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=856,
  serialized_end=920,
)

_STREAMINGRECOGNITIONREQUEST.fields_by_name['config'].message_type = _RECOGNITIONCONFIG
_STREAMINGRECOGNITIONREQUEST.oneofs_by_name['streaming_request'].fields.append(
  _STREAMINGRECOGNITIONREQUEST.fields_by_name['config'])
_STREAMINGRECOGNITIONREQUEST.fields_by_name['config'].containing_oneof = _STREAMINGRECOGNITIONREQUEST.oneofs_by_name['streaming_request']
_STREAMINGRECOGNITIONREQUEST.oneofs_by_name['streaming_request'].fields.append(
  _STREAMINGRECOGNITIONREQUEST.fields_by_name['audio_content'])
_STREAMINGRECOGNITIONREQUEST.fields_by_name['audio_content'].containing_oneof = _STREAMINGRECOGNITIONREQUEST.oneofs_by_name['streaming_request']
_RECOGNITIONCONFIG.fields_by_name['specification'].message_type = _RECOGNITIONSPEC
_RECOGNITIONSPEC.fields_by_name['audio_encoding'].enum_type = _RECOGNITIONSPEC_AUDIOENCODING
_RECOGNITIONSPEC_AUDIOENCODING.containing_type = _RECOGNITIONSPEC
_STREAMINGRECOGNITIONRESPONSE.fields_by_name['chunks'].message_type = _SPEECHRECOGNITIONCHUNK
_SPEECHRECOGNITIONCHUNK.fields_by_name['alternatives'].message_type = _SPEECHRECOGNITIONALTERNATIVE
DESCRIPTOR.message_types_by_name['StreamingRecognitionRequest'] = _STREAMINGRECOGNITIONREQUEST
DESCRIPTOR.message_types_by_name['RecognitionConfig'] = _RECOGNITIONCONFIG
DESCRIPTOR.message_types_by_name['RecognitionSpec'] = _RECOGNITIONSPEC
DESCRIPTOR.message_types_by_name['StreamingRecognitionResponse'] = _STREAMINGRECOGNITIONRESPONSE
DESCRIPTOR.message_types_by_name['SpeechRecognitionChunk'] = _SPEECHRECOGNITIONCHUNK
DESCRIPTOR.message_types_by_name['SpeechRecognitionAlternative'] = _SPEECHRECOGNITIONALTERNATIVE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

StreamingRecognitionRequest = _reflection.GeneratedProtocolMessageType('StreamingRecognitionRequest', (_message.Message,), dict(
  DESCRIPTOR = _STREAMINGRECOGNITIONREQUEST,
  __module__ = 'stt_service_pb2'
  # @@protoc_insertion_point(class_scope:yandex.cloud.ai.stt.v2.StreamingRecognitionRequest)
  ))
_sym_db.RegisterMessage(StreamingRecognitionRequest)

RecognitionConfig = _reflection.GeneratedProtocolMessageType('RecognitionConfig', (_message.Message,), dict(
  DESCRIPTOR = _RECOGNITIONCONFIG,
  __module__ = 'stt_service_pb2'
  # @@protoc_insertion_point(class_scope:yandex.cloud.ai.stt.v2.RecognitionConfig)
  ))
_sym_db.RegisterMessage(RecognitionConfig)

RecognitionSpec = _reflection.GeneratedProtocolMessageType('RecognitionSpec', (_message.Message,), dict(
  DESCRIPTOR = _RECOGNITIONSPEC,
  __module__ = 'stt_service_pb2'
  # @@protoc_insertion_point(class_scope:yandex.cloud.ai.stt.v2.RecognitionSpec)
  ))
_sym_db.RegisterMessage(RecognitionSpec)

StreamingRecognitionResponse = _reflection.GeneratedProtocolMessageType('StreamingRecognitionResponse', (_message.Message,), dict(
  DESCRIPTOR = _STREAMINGRECOGNITIONRESPONSE,
  __module__ = 'stt_service_pb2'
  # @@protoc_insertion_point(class_scope:yandex.cloud.ai.stt.v2.StreamingRecognitionResponse)
  ))
_sym_db.RegisterMessage(StreamingRecognitionResponse)

SpeechRecognitionChunk = _reflection.GeneratedProtocolMessageType('SpeechRecognitionChunk', (_message.Message,), dict(
  DESCRIPTOR = _SPEECHRECOGNITIONCHUNK,
  __module__ = 'stt_service_pb2'
  # @@protoc_insertion_point(class_scope:yandex.cloud.ai.stt.v2.SpeechRecognitionChunk)
  ))
_sym_db.RegisterMessage(SpeechRecognitionChunk)

SpeechRecognitionAlternative = _reflection.GeneratedProtocolMessageType('SpeechRecognitionAlternative', (_message.Message,), dict(
  DESCRIPTOR = _SPEECHRECOGNITIONALTERNATIVE,
  __module__ = 'stt_service_pb2'
  # @@protoc_insertion_point(class_scope:yandex.cloud.ai.stt.v2.SpeechRecognitionAlternative)
  ))
_sym_db.RegisterMessage(SpeechRecognitionAlternative)


DESCRIPTOR._options = None

_STTSERVICE = _descriptor.ServiceDescriptor(
  name='SttService',
  full_name='yandex.cloud.ai.stt.v2.SttService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=923,
  serialized_end=1071,
  methods=[
  _descriptor.MethodDescriptor(
    name='StreamingRecognize',
    full_name='yandex.cloud.ai.stt.v2.SttService.StreamingRecognize',
    index=0,
    containing_service=None,
    input_type=_STREAMINGRECOGNITIONREQUEST,
    output_type=_STREAMINGRECOGNITIONRESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_STTSERVICE)

DESCRIPTOR.services_by_name['SttService'] = _STTSERVICE

# @@protoc_insertion_point(module_scope)



#
# Copyright (c) 2020, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
from types import GeneratorType
from hashlib import sha256
from importlib import import_module
from unidecode import unidecode


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError as e:
        raise ImportError(
            "%s doesn't look like a module path." % dotted_path
        ) from e

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as e:
        raise ImportError(
            'Module "%s" does not define a "%s" attribute/class' % (
                module_path, class_name)
        ) from e


def to_flat_list(data, skip=None, prebool=None, sort=False):
    """
    Преобразовывает многоуровневый объект в [отсортированный] плоский список.
    """
    result = []
    if skip is None:
        skip = ()
    if prebool is None:
        prebool = int

    def collect(item, prefix=''):
        if isinstance(item, dict):
            for key, value in item.items():
                if key in skip:
                    continue
                # Recursive collection for internal values.
                collect(value, prefix='%s%s:' % (prefix, key))
        elif isinstance(item, (list, tuple, range, GeneratorType)):
            for key, value in enumerate(item):
                # Recursive collection for internal values.
                collect(value, prefix='%s%d:' % (prefix, key))
        else:
            if isinstance(item, bool):
                value = prebool(item)
            else:
                value = item
            result.append('%s%s' % (prefix, str(value)))

    collect(data)
    if sort:
        result.sort()
    return result


def to_unique_string(data, skip=None, prebool=None, separator=';'):
    """
    Преобразовывает многоуровневый объект в уникальную строку.
    """
    if data is None:
        return ''
    elif isinstance(data, str):
        return data
    flat_list = to_flat_list(data, skip=skip, prebool=prebool, sort=True)
    return separator.join(flat_list)


def make_checksum(data, *args, **kwargs):
    uniq = to_unique_string(data, *args, **kwargs)
    return sha256(uniq.encode('utf-8')).hexdigest()


def prepare_dialog_name(s):
    """
    Преобразовывает имена разных алфавитов в латинницу и удаляет/заменяет
    лишние символы.
    """
    return unidecode(s).replace(' ', '_').replace('-', '_').replace("'", "")


def levenshtein_distance(text_1, text_2):
    """Алгоритм вычисления дистанции Левенштейна между двумя текстами."""

    length_1, length_2 = len(text_1), len(text_2)
    # Первым должен быть указан наименьший текст.
    if length_1 > length_2:
        text_1, text_2 = text_2, text_1
        length_1, length_2 = length_2, length_1

    current = range(length_1 + 1)
    for i in range(1, length_2 + 1):
        previous, current = current, [i] + [0] * length_1
        for j in range(1, length_1 + 1):
            add = previous[j] + 1
            delete = current[j - 1] + 1
            change = previous[j - 1]
            if text_1[j - 1] != text_2[i - 1]:
                change += 1
            current[j] = min(add, delete, change)
    return current[length_1]



#
# Copyright (c) 2020, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
import json
import re
from datetime import datetime
from gettext import gettext as _
from logging import getLogger
from logging.config import dictConfig
from os import makedirs
from os.path import join as path_join, expanduser, abspath, exists, dirname

from astersay.utils import prepare_dialog_name, import_string

logger = getLogger(__name__)

BASE_DIR = dirname(abspath(__file__))
DEFAULT_WORK_DIR = path_join(expanduser('~'), '.config', 'astersay')
DEFAULT_DIALOG = {
    'constants': {
        'manager_contact': 'Менеджер Василий +79008001234',
    },
    'main_script': [
        # При успехе следующим будет исполнен скрипт '003', иначе '003'.
        {'name': '001', 'success': '002', 'fail': '004'},
        # При любом завершении будет исполнен скрипт '003'.
        {'name': '002'},
        # При любом завершении будет исполнен скрипт '004'.
        {'name': '003'},
        # При любом завершении данные будут переданы указанному плагину.
        # Принимать их или нет - решать плагину.
        {'name': '004', 'plugin': 'example.py'},
    ],
    'scripts': {
        '001': {
            'type': 'parse_names',
            'export_name': 'caller_name',
            'new_buffer': True,
            # Режим чтения буфера с распознанным текстом.
            # 1 - финальный текст, 2 - промежуточный текст, 3 - весь.
            'mode_buffer': 1,
            # Для режимов чтения буфера 2 и 3 следующая переменная определяет
            # начальную итерацию, с которой частичные результаты начнут
            # учитываться.
            # 'partials_begin': 1,
            'say_start': {
                # Вместо текста можно указать файл голоса двумя вариантами:
                # 'voice': '/full/path/to/voice' - без расширения;
                # 'file': '/full/path/to/voicefile.wav' - с расширением.
                'text': _(
                    'Здравствуйте, с Вами говорит базовая диалоговая модель. '
                    'Меня разрабатывали очень весёлые люди. Поэтому я могу '
                    'быть весьма неожиданной. В принципе обо мне - всё. '
                    'Позвольте мне теперь узнать о Вас. Представьтесь, '
                    'пожалуйста.'
                ),
                'escape_digits': '1234567890',
                'nonblocking': True,
            },
            'say_success': {
                'text': _(
                    '%(caller_name)s, мне очень приятно было распознать '
                    'Ваше имя.'
                ),
            },
            'say_fail': {
                'text': _('Увы, но мне не удалось распознать Ваше имя.'),
            },
            'attempts': {
                # Количество итераций.
                'count': 3,
                # Максимальное время одной итерации.
                'iter_time': 15,
                # Размер пауз для ожидания ответа во время одной итерации.
                'iter_pause': 1.0,
                'new_buffer': True,
                'say': {
                    'text': _('Попробуйте назвать ещё раз своё имя.'),
                },
            },
            # Все паузы по-умолчанию равны нулю.
            'pauses': {
                'before': 0,
                'after': 0,
                'before_say_start': 0,
                'after_say_start': 3,
                'before_say_success': 0,
                'after_say_success': 0,
                'before_say_fail': 0,
                'after_say_fail': 0,
                'before_say_attempt': 0,
                'after_say_attempt': 0,
            },
            # Команда для Asterisk - это список аргументов, передаваемых
            # в метод `agi.execute`. Команда выполняется сразу после
            # одноимённой паузы. По-умолчанию команды равны пустому
            # списку и поэтому не производятся.
            'commands': {
                'before': [],
                'after': [],
                'before_say_start': [],
                'after_say_start': [],
                'before_say_success': [],
                'after_say_success': [],
                'before_say_fail': [],
                'after_say_fail': [],
                'before_say_attempt': [],
                'after_say_attempt': [],
            },
        },
        '002': {
            'type': 'search_by_buffer',
            'export_name': 'answered_yes',
            'export_name_text': 'answered_yes_text',
            'new_buffer': True,
            'mode_buffer': 3,
            'partials_begin': 3,
            'regexp': {
                # В такой регулярке отрицания нужно ставить перед
                # положительными ответами.
                'pattern': 'нет|^не | не |заткн|негатив|дура|отрица'
                           'да|конечн|угу|окей|отлично|может|положител',
                'ignorecase': True,
            },
            'booleanize': {
                'true_list': [
                    'да', 'конечн', 'угу', 'окей', 'отлично', 'может',
                    'положител'
                ],
                'text_true': 'положительно',
                'text_false': 'отрицательно'
            },
            'say_start': {
                'text': _(
                    'Если Вы, %(caller_name)s, сейчас скажете какое-нибудь '
                    'положительное или отрицательное утверждение, то я '
                    'попробую распознать и это.'
                ),
                'escape_digits': '1234567890',
                'nonblocking': True,
            },
            'say_success': {
                'text': _('Вы ответили %(answered_yes_text)s, благодарю Вас!.'),
            },
            'say_fail': {
                'text': _('Увы, но мне не удалось распознать утверждение.'),
            },
            'attempts': {
                # Количество итераций.
                'count': 2,
                # Максимальное время одной итерации.
                'iter_time': 10,
                # Размер пауз для ожидания ответа во время одной итерации.
                'iter_pause': 0.5,
                'new_buffer': True,
                'say': {
                    'text': _('Я не поняла Ваш ответ, попробуйте сказать ещё раз.'),
                },
            },
        },
        '003': {
            'type': 'text_processing',
            'file': 'example.py',
            'export_name': 'city',
            'new_buffer': True,
            'mode_buffer': 3,
            'partials_begin': 3,
            'say_start': {
                'text': _(
                    'Пожалуйста, назовите сейчас какой-нибудь известный '
                    'российский город, а я попробую распознать его.'
                ),
                'escape_digits': '1234567890',
                'nonblocking': True,
            },
            'say_success': {
                'text': _('Я распознала %(city)s, благодарю Вас!.'),
            },
            'say_fail': {
                'text': _('Увы, но мне не удалось распознать город.'),
            },
            'attempts': {
                # Количество итераций.
                'count': 3,
                # Максимальное время одной итерации.
                'iter_time': 10,
                # Размер пауз для ожидания ответа во время одной итерации.
                'iter_pause': 0.5,
                'new_buffer': True,
                'say': {
                    'text': _('Я не поняла Ваш ответ, попробуйте сказать ещё раз.'),
                },
            },
        },
        '004': {
            'type': 'simple_say',
            'new_buffer': True,
            'say_start': {
                'text': _('Всего хорошего! До свидания.'),
            }
        },
    }
}


def create_dirs(work_dir):
    dirs = (
        path_join(work_dir, 'config'),
        path_join(work_dir, 'dialogs'),
        path_join(work_dir, 'logs'),
        path_join(work_dir, 'export'),
        path_join(work_dir, 'plugins'),
        path_join(work_dir, 'processors'),
    )
    for directory in dirs:
        makedirs(directory, 0o700, exist_ok=True)
    makedirs(path_join(work_dir, 'storage'), 0o755, exist_ok=True)


def get_dialog_scheme(work_dir, name):
    if not name.lower().endswith('.json'):
        name += '.json'
    filename = path_join(work_dir, 'dialogs', name)
    if not exists(filename):
        if name != 'default.json':
            logger.error(_(
                'Диалог "%s" не найден, будет использоваться диалог '
                '"default.json".'
            ) % name)
        filename = path_join(work_dir, 'dialogs', 'default.json')
        if not exists(filename):
            json.dump(
                DEFAULT_DIALOG, open(filename, 'w'),
                indent=4, ensure_ascii=False)
            logger.info(_('Создан диалог "default.json".'))
            return DEFAULT_DIALOG
    scheme = json.load(open(filename, 'r'))
    return scheme


def get_plugin(work_dir, name):
    filename = path_join(work_dir, 'plugins', name)
    if not exists(filename):
        internal = path_join(BASE_DIR, 'plugins')
        filename = path_join(internal, name)
        if not exists(filename):
            filename = path_join(internal, 'example.py')
    return filename


def get_processor(work_dir, name):
    filename = path_join(work_dir, 'processors', name)
    if not exists(filename):
        internal = path_join(BASE_DIR, 'processors')
        filename = path_join(internal, name)
        if not exists(filename):
            filename = path_join(internal, 'example.py')
    return filename


def get_config(work_dir, default, name):
    filename = path_join(work_dir, 'config', '%s.json' % name)
    if not exists(filename):
        json.dump(default, open(filename, 'w'), indent=4)
    else:
        default.update(json.load(open(filename, 'r')))
    return default


def get_main_config(work_dir):
    default = {
        'recognizer': 'astersay.backends.yandex.YandexRecognizer',
        'synthesizer': 'astersay.backends.yandex.YandexSynthesizer',
        'morphology_processor': 'astersay.morphology.Pymorphy2Processor',
        'export': '{work_dir}/export/{date}/{dialog}_{time}.json',
        'disable_stop_voices': False,
    }
    return get_config(work_dir, default, 'main')


def get_logging_config(work_dir):
    default = {
        'version': 1,
        'disable_existing_loggers': False,
        'filters': {
        },
        'formatters': {
            'verbose': {
                'format': '[%(asctime)s] %(levelname)s <%(name)s>: %(message)s'
            },
            # 'short': {
            #     'format': '%(levelname)s <%(name)s>: %(message)s'
            # },
        },
        'handlers': {
            'main': {
                'class': 'logging.FileHandler',
                'filename': path_join(work_dir, 'logs', 'main.log'),
                'formatter': 'verbose',
            },
            'backend': {
                'class': 'logging.FileHandler',
                'filename': path_join(work_dir, 'logs', 'backend.log'),
                'formatter': 'verbose',
            },
            'dialog': {
                'class': 'logging.FileHandler',
                'filename': path_join(work_dir, 'logs', 'dialog.log'),
                'formatter': 'verbose',
            },
        },
        'loggers': {
            'astersay': {
                'handlers': ['main'],
                'level': 'INFO',
            },
            'astersay.backends': {
                'handlers': ['backend'],
                'level': 'DEBUG',
            },
            'astersay.agi': {
                'handlers': ['dialog'],
                'level': 'DEBUG',
            },
            'astersay.dialog': {
                'handlers': ['dialog'],
                'level': 'DEBUG',
            },
            'astersay.conf': {
                'handlers': ['dialog'],
                'level': 'DEBUG',
            },
            'astersay.backends.yandex': {
                'handlers': ['dialog'],
                'level': 'INFO',
            },
            'astersay.backends.vosk': {
                'handlers': ['dialog'],
                'level': 'INFO',
            },
        }
    }
    return get_config(work_dir, default, 'logging')


def get_yandex_config(work_dir):
    default = {
        'auth': {
            'private_filename': path_join(work_dir, 'yandex_private.pem'),
            'token_filename': path_join(work_dir, 'yandex_token.json'),
            'key_id': '',
            'service_account_id': '',
            'folder_id': '',
        },
        'recognizer': {
            'model': 'general',
            'audio_encoding': 'LINEAR16_PCM',
            'chunk_size': 1024,
            'language_code': 'ru-RU',
            'partial_results': True,
            'profanity_filter': False,
            'single_utterance': False,
            'raw_results': False,
            'sample_rate_hertz': 8000,
        },
        'synthesizer': {
            'emotion': 'neutral',
            'format': 'lpcm',
            'lang': 'ru-RU',
            'sample_rate_hertz': 48000,
            'speed': 1.0,
            'storage': path_join(work_dir, 'storage'),
            'voice': 'oksana',
            'convertor': {
                'command': (
                    'sox -t raw -r 48k -b 16 -e signed-integer -c 1 "%s" '
                    '-t wav -r 8k -c 1 "%s"'
                ),
                'src_format': 'raw',
                'src_delete': True,
            }
        },
    }
    return get_config(work_dir, default, 'yandex')


def get_vosk_config(work_dir):
    default = {
        'connection': {
            'host': 'localhost',
            'port': 2700,
        },
        'recognizer': {
            'chunk_size': 8000,
            'max_size': 10485760,
            'max_time': 300,
            'sample_rate_hertz': 8000,
        },
    }
    return get_config(work_dir, default, 'vosk')


class Settings:
    """
    Конструктор для объекта управления настройками запускаемого приложения.
    """

    work_dir = None
    dialog_name = 'default'
    main = None
    logging = None
    yandex = None
    vosk = None

    def configure(self):
        self.work_dir = work_dir = abspath(self.work_dir or DEFAULT_WORK_DIR)
        create_dirs(work_dir)

        self.main = get_main_config(work_dir)
        self.logging = get_logging_config(work_dir)
        self.yandex = get_yandex_config(work_dir)
        self.vosk = get_vosk_config(work_dir)

        self.reconfigure_logging()

    def reconfigure_logging(self, logging=None):
        logging = logging or self.logging
        root = getLogger()
        map(root.removeHandler, root.handlers[:])
        map(root.removeFilter, root.filters[:])
        dictConfig(logging)

    @property
    def dialog_scheme(self):
        """Динамическая загрузка внешних изменений модели диалога."""
        return get_dialog_scheme(self.work_dir, self.dialog_name)

    def set_dialog_name(self, name):
        name = prepare_dialog_name(name or '')
        if name:
            self.dialog_name = name
        return self.dialog_name

    def get_recognizer(self):
        if not hasattr(self, '_recognizer'):
            Recognizer = import_string(self.main['recognizer'])
            self._recognizer = Recognizer(self)
        return self._recognizer

    def get_synthesizer(self):
        if not hasattr(self, '_synthesizer'):
            Synthesizer = import_string(self.main['synthesizer'])
            self._synthesizer = Synthesizer(self)
        return self._synthesizer

    def get_morphology_processor(self):
        module = self.main['morphology_processor']
        try:
            Processor = import_string(module)
        except ImportError as e:
            logger.error(_('Модуль %s не установлен.'), module, exc_info=e)
            Processor = import_string('astersay.morphology.Pymorphy2Processor')
        return Processor()

    def get_morphology_pattern(self):
        if self.main['morphology']:
            pattern = self.main['morphology_pattern']
            if pattern:
                return re.compile(pattern)
        return None

    def get_plugin(self, name):
        """
        Возвращает полный путь к гарантированно существующему файлу плагина.
        Когда плагина нет, то вернёт путь к файлу примера.
        """
        if name.startswith('/') and exists(name):
            return name
        return get_plugin(self.work_dir, name)

    def get_processor(self, name):
        """
        Возвращает полный путь к гарантированно существующему файлу внешнего
        обработчика. Когда такого обработчика нет, то вернёт путь к файлу
        примера.
        """
        if name.startswith('/') and exists(name):
            return name
        return get_processor(self.work_dir, name)

    def export_dialog(self, agi_params_data: dict, export_data: dict,
                      text_buffers: dict, **extra):
        export = self.main.get('export')
        if not export:
            return

        now = datetime.now()
        date = now.date()
        time = now.time()
        kw = {
            'work_dir': self.work_dir,
            'dialog': self.dialog_name,
            'now': now.isoformat(),
            'date': date.isoformat(),
            'time': time.isoformat(),
            'year': date.year,
            'month': date.month,
            'day': date.day,
        }
        filename = export.format(**kw)
        makedirs(dirname(filename), 0o700, exist_ok=True)
        data = {
            'agi': agi_params_data,
            'export': export_data,
            'speech': text_buffers,
        }
        if extra:
            assert 'agi' not in extra
            assert 'export' not in extra
            assert 'speech' not in extra
            data.update(extra)
        json.dump(data, open(filename, 'w'),
                  indent=4, ensure_ascii=False)
        return filename

    def get_stop_file(self):
        if self.main.get('disable_stop_voices'):
            logger.info(_('Остановка голоса отключена в настройках.'))
            return ''
        filename = path_join(BASE_DIR, 'sounds', 'stop_voice_file.wav')
        return filename if exists(filename) else ''


settings = Settings()
# Удаляем конструктор - он больше не нужен.
del Settings



#
# Copyright (c) 2020, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#


class AstersayError(Exception):
    pass


class AgiError(AstersayError):
    pass


class AgiUnknownError(AgiError):
    pass


class AgiAppError(AgiError):
    pass


class AgiHangupError(AgiAppError):
    pass


class AgiSighupError(AgiHangupError):
    pass


class AgiSigpipeError(AgiHangupError):
    pass


class AgiUsageError(AgiError):
    pass


class AgiInvalidCommand(AgiError):
    pass


class AppError(AstersayError):
    pass


class ImproperlyConfigured(AstersayError):
    pass


class TokenError(AstersayError):
    pass
